import unittest
import numpy as np

# VTK imports:
import vtk
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa
from .. import _helpers

# Functionality to test:
from .poly import *
from .slicing import *
from .tables import *
from .voxelize import *
from .xyz import *

RTOL = 0.000001


###############################################################################
###############################################################################

class TestCombineTables(unittest.TestCase):
    """
    Test the `CombineTables` filter
    """
    def setUp(self):
        # Create some input tables
        self.t0 = vtk.vtkTable()
        self.t1 = vtk.vtkTable()
        # Populate the tables
        self.n = 100
        self.titles = ('Array 0', 'Array 1', 'Array 2')
        self.arrs = [None, None, None]
        self.arrs[0] = np.random.random(self.n) # Table 0
        self.arrs[1] = np.random.random(self.n) # Table 0
        self.arrs[2] = np.random.random(self.n) # Table 1
        self.t0.AddColumn(_helpers.numToVTK(self.arrs[0], self.titles[0]))
        self.t0.AddColumn(_helpers.numToVTK(self.arrs[1], self.titles[1]))
        self.t1.AddColumn(_helpers.numToVTK(self.arrs[2], self.titles[2]))
        # Now use the `CombineTables` filter:
        f = CombineTables()
        f.SetInputDataObject(0, self.t0)
        f.SetInputDataObject(1, self.t1)
        f.Update()
        self.TABLE = f.GetOutputDataObject(0)


    #########################


    def test_shape(self):
        """`CombineTables`: table shape"""
        self.assertEqual(self.TABLE.GetNumberOfColumns(), len(self.titles))
        self.assertEqual(self.TABLE.GetNumberOfRows(), self.n)

    def test_data_array_names(self):
        """`CombineTables`: data array names"""
        for i in range(len(self.titles)):
            self.assertEqual(self.TABLE.GetColumnName(i), self.titles[i])

    def test_data_fidelity(self):
        """`CombineTables`: data fidelity"""
        wpdi = dsa.WrapDataObject(self.TABLE)
        for i in range(len(self.titles)):
            arr = wpdi.RowData[self.titles[i]]
            self.assertTrue(np.allclose(arr, self.arrs[i], rtol=RTOL))

###############################################################################


class TestReshapeTable(unittest.TestCase):
    """
    Test the `ReshapeTable` filter
    """
    def setUp(self):
        # Create some input tables
        self.t0 = vtk.vtkTable()
        # Populate the tables
        self.arrs = [None, None, None]
        self.n = 400
        self.ncols = 2
        self.nrows = int(self.n * len(self.arrs) / self.ncols)
        self.titles = ('Array 0', 'Array 1', 'Array 2')
        self.arrs[0] = np.random.random(self.n) # Table 0
        self.arrs[1] = np.random.random(self.n) # Table 0
        self.arrs[2] = np.random.random(self.n) # Table 1
        self.t0.AddColumn(_helpers.numToVTK(self.arrs[0], self.titles[0]))
        self.t0.AddColumn(_helpers.numToVTK(self.arrs[1], self.titles[1]))
        self.t0.AddColumn(_helpers.numToVTK(self.arrs[2], self.titles[2]))
        return


    def _check_shape(self, table):
        self.assertEqual(table.GetNumberOfRows(), self.nrows)
        self.assertEqual(table.GetNumberOfColumns(), self.ncols)
        return

    def _check_data_fidelity(self, table, order):
        wpdi = dsa.WrapDataObject(table)
        tarr = np.zeros((self.nrows, self.ncols))
        for i in range(self.ncols):
            tarr[:,i] = wpdi.RowData[i]
        arrs = np.array(self.arrs).T
        arrs = arrs.flatten()
        arrs = np.reshape(arrs, (self.nrows, self.ncols), order=order)
        self.assertEqual(tarr.shape, arrs.shape)
        self.assertTrue(np.allclose(tarr, arrs, rtol=RTOL))
        return

    def _check_data_array_titles(self, table, titles):
        for i in range(len(titles)):
            self.assertEqual(table.GetColumnName(i), titles[i])
        return

    def _generate_output(self, order, titles=None):
        f = ReshapeTable()
        f.SetInputDataObject(0, self.t0)
        f.SetNumberOfColumns(self.ncols)
        f.SetNumberOfRows(self.nrows)
        f.SetOrder(order)
        if titles is not None:
            f.SetNames(titles)
        f.Update()
        return f.GetOutputDataObject(0)

    ###############

    def test_reshape_f(self):
        """`ReshapeTable`: F-order, no input names"""
        order = 'F'
        table = self._generate_output(order, titles=None)
        # Check output:
        self._check_shape(table)
        self._check_data_fidelity(table, order)
        self._check_data_array_titles(table, ['Field %d' % i for i in range(self.ncols)])
        return

    def test_reshape_f_names(self):
        """`ReshapeTable`: F-order, input names given"""
        order = 'F'
        titles = ['Title %d' % i for i in range(self.ncols)]
        table = self._generate_output(order, titles=titles)
        # Check output:
        self._check_shape(table)
        self._check_data_fidelity(table, order)
        self._check_data_array_titles(table, titles)
        return


    def test_reshape_c(self):
        """`ReshapeTable`: C-order, input names given as string"""
        order = 'C'
        titles = ['Title %d' % i for i in range(self.ncols)]
        ts = ';'.join(t for t in titles)
        table = self._generate_output(order, titles=ts)
        # Check output:
        self._check_shape(table)
        self._check_data_fidelity(table, order)
        self._check_data_array_titles(table, titles)
        return

    def test_reshape_c_names(self):
        """`ReshapeTable`: C-order, few input names given"""
        order = 'C'
        fewtitles = ['Title %d' % i for i in range(self.ncols - 2)]
        rest = ['Field %d' % i for i in range(2)]
        table = self._generate_output(order, titles=fewtitles)
        # Check output:
        self._check_shape(table)
        self._check_data_fidelity(table, order)
        self._check_data_array_titles(table, fewtitles + rest)
        return




###############################################################################

ROTATED_TEXT = """326819.497,4407450.636,1287.5
326834.34,4407470.753,1287.5
326849.183,4407490.87,1287.5
326864.027,4407510.986,1287.5
326878.87,4407531.103,1287.5
326893.713,4407551.22,1287.5
326908.556,4407571.336,1287.5
326923.399,4407591.453,1287.5
326938.242,4407611.57,1287.5
326953.086,4407631.686,1287.5"""

ROTATED_POINTS = np.genfromtxt((line.encode('utf8') for line in ROTATED_TEXT.split('\n')), delimiter=',''', dtype=float)

class TestRotationTool(unittest.TestCase):
    """
    Test the `RotationTool` filter
    """
    # An example voxel:
    # voxel = np.array([
    #     [0,0,0],
    #     [0,0,1],
    #     [0,1,1],
    #     [1,1,1],
    #     [0,1,0],
    #     [1,0,0],
    #     [1,1,0],
    # ])

    def setUp(self):
        self.RTOL = 0.00001 # As higi as rotation precision can get
        return


    def test_recovery(self):
        """`RotationTool`: Test a simple rotation recovery"""
        r = RotationTool()
        # Input points
        x = np.array([1.1, 1.1, 1.1, 2.1, 2.1, 2.1])
        y = np.array([1.0, 2.0, 3.0, 1.0, 2.0, 3.0])
        z = np.zeros(len(x))
        x = np.reshape(x, (len(x), -1))
        y = np.reshape(y, (len(y), -1))
        z = np.reshape(z, (len(z), -1))
        pts = np.concatenate((x,y,z), axis=1)
        rot = np.deg2rad(-33.3)
        pts[:, 0:2] = r.Rotate(pts[:, 0:2], rot)
        xx, yy, zz, dx, dy, angle = r.EstimateAndRotate(pts[:,0], pts[:,1], pts[:,2])
        rpts = np.vstack((xx,yy,zz)).T
        self.assertTrue(np.allclose(angle, np.deg2rad(33.3), rtol=RTOL), msg='Recovered angle is incorrect.')
        self.assertTrue(np.allclose(dx, 1.0, rtol=RTOL), msg='Recovered x-spacing is incorrect.')
        self.assertTrue(np.allclose(dy, 1.0, rtol=RTOL), msg='Recovered y-spacing is incorrect.')
        # Now check coordinates...
        self.assertTrue(np.allclose(rpts, np.concatenate((x,y,z), axis=1), rtol=self.RTOL), msg='Recovered coordinates are incorrect.')
        return

    def test_bradys(self):
        """`RotationTool`: This is primarily to make sure no errors arise"""
        r = RotationTool()
        pts = ROTATED_POINTS
        dx, dy, angle = r.EstimateAndRotate(pts[:,0], pts[:,1], pts[:,2])[3::]
        self.assertTrue(np.allclose(angle, np.deg2rad(53.55), rtol=self.RTOL), msg='Recovered angle is incorrect.')
        self.assertTrue(np.allclose(dx, 25.0, rtol=self.RTOL), msg='Recovered x-spacing is incorrect.')
        self.assertTrue(np.allclose(dy, 25.0, rtol=self.RTOL), msg='Recovered y-spacing is incorrect.')
        return



class TestRotatePoints(unittest.TestCase):
    """
    Test the `RotatePoints` filter
    """
    def setUp(self):
        self.RTOL = 0.00001 # As higi as rotation precision can get
        x = np.array([0.0,1.0,0.0])
        y = np.array([0.0,0.0,1.0])
        z = np.array([0.0,0.0,0.0])
        x = np.reshape(x, (len(x), -1))
        y = np.reshape(y, (len(y), -1))
        z = np.reshape(z, (len(z), -1))
        self.pts = np.concatenate((x,y,z), axis=1)
        self.vtkpoints = PointsToPolyData(self.pts)
        return

    def test_rotation(self):
        f = RotatePoints()
        f.SetInputDataObject(self.vtkpoints)
        f.SetRotationDegrees(33.3)
        f.Update()
        output = f.GetOutput()
        self.assertIsNotNone(output)


###############################################################################


class TestVoxelizePoints(unittest.TestCase):
    """
    Test the `VoxelizePoints` filter
    """

    def test_simple_case(self):
        """`VoxelizePoints`: simple case"""
        x = np.array([0.0,1.0,0.0])
        y = np.array([0.0,0.0,1.0])
        z = np.array([0.0,0.0,0.0])
        x = np.reshape(x, (len(x), -1))
        y = np.reshape(y, (len(y), -1))
        z = np.reshape(z, (len(z), -1))
        pts = np.concatenate((x,y,z), axis=1)
        vtkpoints = PointsToPolyData(pts)
        # Use filter
        v = VoxelizePoints()
        v.SetInputDataObject(vtkpoints)
        v.SetSafeSize(5.0)
        v.Update()
        grid = v.GetOutput()
        # Checkout output:
        self.assertEqual(grid.GetNumberOfCells(), 3, msg='Number of CELLS is incorrect')
        self.assertEqual(grid.GetNumberOfPoints(), 16, msg='Number of POINTS is incorrect')
        bounds = grid.GetBounds()
        self.assertEqual(bounds, (-0.5,1.5, -0.5,1.5, -2.5,2.5), msg='Grid bounds are incorrect.') # Z bounds from SAFE
        return


    def test_simple_rotated_case(self):
        """`VoxelizePoints`: simple rotated case"""
        pts = ROTATED_POINTS
        vtkpoints = PointsToPolyData(ROTATED_POINTS)
        # Use filter
        v = VoxelizePoints()
        v.SetInputDataObject(vtkpoints)
        v.SetSafeSize(5.0)
        v.Update()
        grid = v.GetOutput()
        # Checkout output:
        #- Assumes this same data's rotation was checked by `TestRotationTool`
        self.assertEqual(grid.GetNumberOfCells(), len(pts), msg='Number of CELLS is incorrect')
        numPts = (len(pts) * 8) - ((len(pts) - 1) * 4) # Works because points make a line
        self.assertEqual(grid.GetNumberOfPoints(), numPts, msg='Number of POINTS is incorrect')
        return

    def test_mesh_grid_uniform(self):
        """`VoxelizePoints`: uniform mesh grid with given spacings"""
        # make the mesh grid
        dd = 5
        x = y = z = np.arange(0, 100, dd, dtype=float)
        g = np.meshgrid(x, y, z)
        # Convert to XYZ points
        points = np.vstack(map(np.ravel, g)).T
        rand = np.random.random(len(points))
        vtkpoints = PointsToPolyData(points)
        vtkpoints.GetPointData().AddArray(_helpers.numToVTK(rand, 'Random'))
        # Use filter
        v = VoxelizePoints()
        v.SetInputDataObject(vtkpoints)
        v.SetEstimateGrid(False) # Cell size is explicitly set
        v.SetDeltaX(10)
        v.SetDeltaY(10)
        v.SetDeltaZ(10)
        v.Update()
        grid = v.GetOutput()
        wgrd = dsa.WrapDataObject(grid)
        celldata = wgrd.CellData['Random']
        # Checkout output:
        self.assertEqual(grid.GetNumberOfCells(), 8*10**3, msg='Number of CELLS is incorrect')
        numPts = (len(x)+2)**3
        self.assertEqual(grid.GetNumberOfPoints(), numPts, msg='Number of POINTS is incorrect')
        self.assertTrue(np.allclose(celldata, rand))

        # Now check that we can set the spacing for every cell
        spac = np.full((len(points)), 10.0)
        v.SetDeltas(spac, spac, spac)
        v.Update()
        grid = v.GetOutput()
        wgrd = dsa.WrapDataObject(grid)
        celldata = wgrd.CellData['Random']
        self.assertEqual(grid.GetNumberOfCells(), 8*10**3, msg='Number of CELLS is incorrect')
        self.assertEqual(grid.GetNumberOfPoints(), numPts, msg='Number of POINTS is incorrect')
        self.assertTrue(np.allclose(celldata, rand))
        return




###############################################################################


class TestExtractPoints(unittest.TestCase):
    """
    Test the `ExtractPoints` filter
    """
    def test_bad_extraction(self):
        """`ExtractPoints`: catch a bad extraction"""
        img = vtk.vtkImageData()
        img.SetDimensions(10, 10, 10)
        f = ExtractPoints()
        f.SetInputDataObject(img)
        f.Update()
        self.assertTrue(f.ErrorOccurred())
        return



###############################################################################

class TestArrayMath(unittest.TestCase):
    """
    Test the `ArrayMath` filter
    """

    def setUp(self):
        # Create some input tables
        self.t0 = vtk.vtkTable()
        # Populate the tables
        self.arrs = [None, None]
        self.n = 400
        self.titles = ('Array 0', 'Array 1')
        self.arrs[0] = np.random.random(self.n) # Table 0
        self.arrs[1] = np.random.random(self.n) # Table 0
        self.t0.AddColumn(_helpers.numToVTK(self.arrs[0], self.titles[0]))
        self.t0.AddColumn(_helpers.numToVTK(self.arrs[1], self.titles[1]))
        return

    def test_get_operations(self):
        op = ArrayMath.GetOperation('add')
        self.assertIsNotNone(op)
        op = ArrayMath.GetOperation('subtract')
        self.assertIsNotNone(op)
        op = ArrayMath.GetOperation('multiply')
        self.assertIsNotNone(op)
        op = ArrayMath.GetOperation('divide')
        self.assertIsNotNone(op)
        op = ArrayMath.GetOperation('correlate')
        self.assertIsNotNone(op)

    def _gen_and_check(self, op, check, flip=False):
        # Perform filter
        f = ArrayMath()
        f.SetInputDataObject(self.t0)
        if flip:
            f.SetInputArrayToProcess(1, 0, 0, 6, self.titles[0]) # field 6 is row data
            f.SetInputArrayToProcess(0, 0, 0, 6, self.titles[1]) # field 6 is row data
        else:
            f.SetInputArrayToProcess(0, 0, 0, 6, self.titles[0]) # field 6 is row data
            f.SetInputArrayToProcess(1, 0, 0, 6, self.titles[1]) # field 6 is row data
        f.SetOperation(op)
        f.SetNewArrayName('test')
        f.Update()
        # Now test the result
        output = f.GetOutput()
        wout = dsa.WrapDataObject(output)
        arr = wout.RowData['test']
        self.assertTrue(np.allclose(arr, check, rtol=RTOL))


    def test_add(self):
        op = ArrayMath.GetOperation('add')
        check = self.arrs[0] + self.arrs[1]
        self._gen_and_check(op, check)
        # now flip order and check
        # result should be same
        self._gen_and_check(op, check, flip=True)


    def test_subtract(self):
        op = ArrayMath.GetOperation('subtract')
        check = self.arrs[0] - self.arrs[1]
        self._gen_and_check(op, check)
        # now flip order and check
        check = self.arrs[1] - self.arrs[0]
        self._gen_and_check(op, check, flip=True)

    def test_multiply(self):
        op = ArrayMath.GetOperation('multiply')
        check = self.arrs[0] * self.arrs[1]
        self._gen_and_check(op, check)
        # now flip order and check
        # result should be same
        self._gen_and_check(op, check, flip=True)

    def test_divide(self):
        op = ArrayMath.GetOperation('divide')
        check = self.arrs[0] / self.arrs[1]
        self._gen_and_check(op, check)
        # now flip order and check
        check = self.arrs[1] / self.arrs[0]
        self._gen_and_check(op, check, flip=True)

    def test_correlate(self):
        op = ArrayMath.GetOperation('correlate')
        check = np.correlate(self.arrs[0], self.arrs[1], mode='same')
        self._gen_and_check(op, check)
        # now flip order and check
        check = np.correlate(self.arrs[1], self.arrs[0], mode='same')
        self._gen_and_check(op, check, flip=True)



###############################################################################

class TestNormalizeArray(unittest.TestCase):
    """
    Test the `NormalizeArray` filter
    """

    def setUp(self):
        # Create some input tables
        self.t0 = vtk.vtkTable()
        # Populate the tables
        self.n = 400
        self.title = 'Array 0'
        self.arr = np.random.random(self.n) # Table 0
        self.t0.AddColumn(_helpers.numToVTK(self.arr, self.title))
        return

    def test_get_operations(self):
        op = NormalizeArray.GetNormalization('feature_scale')
        self.assertIsNotNone(op)
        op = NormalizeArray.GetNormalization('standard_score')
        self.assertIsNotNone(op)
        op = NormalizeArray.GetNormalization('log10')
        self.assertIsNotNone(op)
        op = NormalizeArray.GetNormalization('natural_log')
        self.assertIsNotNone(op)
        op = NormalizeArray.GetNormalization('just_multiply')
        self.assertIsNotNone(op)

    def _gen_and_check(self, op, check, flip=False):
        # Perform filter
        f = NormalizeArray()
        f.SetInputDataObject(self.t0)
        f.SetInputArrayToProcess(0, 0, 0, 6, self.title) # field 6 is row data
        f.SetNormalization(op)
        f.SetNewArrayName('test')
        f.Update()
        # Now test the result
        output = f.GetOutput()
        wout = dsa.WrapDataObject(output)
        arr = wout.RowData['test']
        self.assertTrue(np.allclose(arr, check, rtol=RTOL))


    def test_feature_scale(self):
        op = NormalizeArray.GetNormalization('feature_scale')
        check = NormalizeArray._featureScale(self.arr)
        self._gen_and_check(op, check)

    def test_standard_score(self):
        op = NormalizeArray.GetNormalization('standard_score')
        check = NormalizeArray._standardScore(self.arr)
        self._gen_and_check(op, check)

    def test_log10(self):
        op = NormalizeArray.GetNormalization('log10')
        check = NormalizeArray._log10(self.arr)
        self._gen_and_check(op, check)

    def test_natural_log(self):
        op = NormalizeArray.GetNormalization('natural_log')
        check = NormalizeArray._logNat(self.arr)
        self._gen_and_check(op, check)

    def test_just_multiply(self):
        op = NormalizeArray.GetNormalization('just_multiply')
        check = NormalizeArray._passArray(self.arr)
        self._gen_and_check(op, check)


###############################################################################

class TestAddCellConnToPoints(unittest.TestCase):
    """
    Test the `AddCellConnToPoints` filter
    """

    def makeSimpleInput(self):
        x = np.array([0.0,1.0,0.0])
        y = np.array([0.0,0.0,1.0])
        z = np.array([0.0,0.0,0.0])
        x = np.reshape(x, (len(x), -1))
        y = np.reshape(y, (len(y), -1))
        z = np.reshape(z, (len(z), -1))
        self.pts = np.concatenate((x,y,z), axis=1)
        self.vtkpoints = PointsToPolyData(self.pts)

    def makeComplicatedInput(self, shuffle=True):
        def path1(y):
            # Equation: x = a(y-h)^2 + k
            k = 110.0
            h = 0.0
            a = - k / 160.0**2
            x = a*(y-h)**2 + k
            idxs = np.argwhere(x>0)
            return x[idxs][:,0], y[idxs][:,0]

        y = np.arange(0.0,10.0)
        zo = np.linspace(9.0,11.0, num=len(y))
        x,y = path1(y)

        coords = np.zeros((len(y),3))
        coords[:,0] = x
        coords[:,1] = y
        coords[:,2] = zo

        np.random.shuffle(coords)
        self.pts = coords
        self.vtkpoints = PointsToPolyData(self.pts)

    def test_poly_line(self):
        self.makeSimpleInput()
        f = AddCellConnToPoints()
        f.SetInputDataObject(self.vtkpoints)
        f.SetCellType(vtk.VTK_POLY_LINE)
        f.Update()
        output = f.GetOutput()
        self.assertEqual(1, output.GetNumberOfCells())
        # Now test nearest neighbor functionality
        self.makeComplicatedInput()
        f = AddCellConnToPoints()
        f.SetInputDataObject(self.vtkpoints)
        f.SetCellType(vtk.VTK_POLY_LINE)
        f.SetUseNearestNbr(True)
        f.Update()
        output = f.GetOutput()
        self.assertEqual(1, output.GetNumberOfCells())
        # Its fairly difficult to test the nearest neighbor approximations...
        # This was done visually in ParaView.
        # The above test is just there to make sure no errors are thrown
        # NOTE: assumes developers visually inspect in ParaView if functionality changes
        return

    def test_line(self):
        self.makeSimpleInput()
        f = AddCellConnToPoints()
        f.SetInputDataObject(self.vtkpoints)
        f.SetCellType(vtk.VTK_LINE)
        f.Update()
        output = f.GetOutput()
        self.assertEqual(len(self.pts)-1, output.GetNumberOfCells())
        # Now test nearest neighbor functionality
        self.makeComplicatedInput()
        f = AddCellConnToPoints()
        f.SetInputDataObject(self.vtkpoints)
        f.SetCellType(vtk.VTK_LINE)
        f.SetUseNearestNbr(True)
        f.Update()
        output = f.GetOutput()
        self.assertEqual(len(self.pts)-1, output.GetNumberOfCells())
        # Its fairly difficult to test the nearest neighbor approximations...
        # This was done visually in ParaView.
        # The above test is just there to make sure no errors are thrown
        # NOTE: assumes developers visually inspect in ParaView if functionality changes
        return


###############################################################################

class TestPointsToTube(unittest.TestCase):
    """
    Test the `PointsToTube` filter
    """
    def makeComplicatedInput(self, shuffle=True):
        def path1(y):
            # Equation: x = a(y-h)^2 + k
            k = 110.0
            h = 0.0
            a = - k / 160.0**2
            x = a*(y-h)**2 + k
            idxs = np.argwhere(x>0)
            return x[idxs][:,0], y[idxs][:,0]

        y = np.linspace(0.0, 200.0, num=100)
        x,y = path1(y)
        zo = np.linspace(9.0,11.0, num=len(y))

        coords = np.zeros((len(y),3))
        coords[:,0] = x
        coords[:,1] = y
        coords[:,2] = zo

        np.random.shuffle(coords)
        self.pts = coords
        self.vtkpoints = PointsToPolyData(self.pts)

    def test_(self):
        self.makeComplicatedInput()
        f = PointsToTube()
        f.SetInputDataObject(self.vtkpoints)
        f.SetRadius(20)
        f.SetNumberOfSides(10)
        f.SetUseNearestNbr(True)
        f.Update()
        output = f.GetOutput()
        self.assertEqual(10, output.GetNumberOfCells())
        self.assertEqual(10*len(self.pts), output.GetNumberOfPoints())


# ###############################################################################
#
#
# class TestManySlicesAlongPoints(unittest.TestCase):
#     """
#     Test the `ManySlicesAlongPoints` filter
#     """
#
#     def test_(self):
#         self.assertTrue(False)
#
#
#
# ###############################################################################
#
# class TestManySlicesAlongAxis(unittest.TestCase):
#     """
#     Test the `ManySlicesAlongAxis` filter
#     """
#
#     def test_(self):
#         self.assertTrue(False)
#
#
# ###############################################################################
#
# class TestSliceThroughTime(unittest.TestCase):
#     """
#     Test the `SliceThroughTime` filter
#     """
#
#     def test_(self):
#         self.assertTrue(False)
#
#
# ###############################################################################
#
# class TestExtractPoints(unittest.TestCase):
#     """
#     Test the `ExtractPoints` filter
#     """
#
#     def test_(self):
#         self.assertTrue(False)
#
#
# ###############################################################################



__all__ = [
    'TranslateGridOrigin'
]

import vtk
from vtk.util import numpy_support as nps
import numpy as np
# Import Helpers:
from ..base import FilterBase
from .. import _helpers


#---- Translate Grid Origin ----#

class TranslateGridOrigin(FilterBase):
    """This filter will translate the origin of `vtkImageData` to any specified Corner of the data set assuming it is currently in the South West Bottom Corner (will not work if Corner was moved prior).
    """
    __displayname__ = 'Translate Grid Origin'
    __type__ = 'filter'
    def __init__(self, corner=1):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkImageData',
            nOutputPorts=1, outputType='vtkImageData')
        self.__corner = corner


    def _Translate(self, pdi, pdo):
        if pdo is None:
            pdo = vtk.vtkImageData()

        [nx, ny, nz] = pdi.GetDimensions()
        [sx, sy, sz] = pdi.GetSpacing()
        [ox, oy, oz] = pdi.GetOrigin()

        pdo.DeepCopy(pdi)

        xx,yy,zz = 0.0,0.0,0.0

        if self.__corner == 1:
            # South East Bottom
            xx = ox - (nx-1)*sx
            yy = oy
            zz = oz
        elif self.__corner == 2:
            # North West Bottom
            xx = ox
            yy = oy - (ny-1)*sy
            zz = oz
        elif self.__corner == 3:
            # North East Bottom
            xx = ox - (nx-1)*sx
            yy = oy - (ny-1)*sy
            zz = oz
        elif self.__corner == 4:
            # South West Top
            xx = ox
            yy = oy
            zz = oz - (nz-1)*sz
        elif self.__corner == 5:
            # South East Top
            xx = ox - (nx-1)*sx
            yy = oy
            zz = oz - (nz-1)*sz
        elif self.__corner == 6:
            # North West Top
            xx = ox
            yy = oy - (ny-1)*sy
            zz = oz - (nz-1)*sz
        elif self.__corner == 7:
            # North East Top
            xx = ox - (nx-1)*sx
            yy = oy - (ny-1)*sy
            zz = oz - (nz-1)*sz

        pdo.SetOrigin(xx, yy, zz)

        return pdo

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output.
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._Translate(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def SetCorner(self, corner):
        """Set the corner to use

        Args:
            corner (int) : corner location; see note.

        Note:
            * 1: South East Bottom
            * 2: North West Bottom
            * 3: North East Bottom
            * 4: South West Top
            * 5: South East Top
            * 6: North West Top
            * 7: North East Top
        """
        if self.__corner != corner:
            self.__corner = corner
            self.Modified()



__all__ = [
    'PackedBinariesReader',
    'MadagascarReader']

import numpy as np
from vtk.util import numpy_support as nps
import vtk
import warnings
# Import Helpers:
from .. import _helpers
from ..base import ReaderBase



class PackedBinariesReader(ReaderBase):
    """This reads in float or double data that is packed into a binary file format. It will treat the data as one long array and make a ``vtkTable`` with one column of that data. The reader uses defaults to import as floats with native endianness. Use the Table to Uniform Grid or the Reshape Table filters to give more meaning to the data. We chose to use a ``vtkTable`` object as the output of this reader because it gives us more flexibility in the filters we can apply to this data down the pipeline and keeps thing simple when using filters in this repository.
    """
    __displayname__ = 'Packed Binaries Reader'
    __type__ = 'reader'
    def __init__(self, **kwargs):
        ReaderBase.__init__(self,
            nOutputPorts=1, outputType='vtkTable', **kwargs)
        # Other Parameters
        self.__dataName = kwargs.get('dataname', 'Data')
        self.__dtypechar = kwargs.get('dtype', 'f')
        self.__endian = kwargs.get('endian', '')
        self.__dtype, self.__vtktype = _helpers.getdTypes(dtype=self.__dtypechar, endian=self.__endian)
        # Data objects to hold the read data for access by the pipeline methods
        self.__data = []

    def _ReadRawFile(self, fileName):
        dtype = self.__dtype
        if dtype == np.dtype('>f'):
            # Checks if big-endian and fixes read
            dtype = np.dtype('f')
        try:
            arr = np.fromfile(fileName, dtype=dtype)
        except (FileNotFoundError, OSError) as fe:
            raise _helpers.PVGeoError(str(fe))
        return np.asarray(arr, dtype=self.__dtype)

    def _GetFileContents(self, idx=None):
        if idx is not None:
            fileNames = [self.GetFileNames(idx=idx)]
        else:
            fileNames = self.GetFileNames()
        contents = []
        for f in fileNames:
            contents.append(self._ReadRawFile(f))
        if idx is not None: return contents[0]
        return contents


    def _ReadUpFront(self):
        """Should not need to be overridden
        """
        # Perform Read
        self.__data = self._GetFileContents()
        self.NeedToRead(flag=False)
        return 1

    def _GetRawData(self, idx=0):
        """This will return the proper data for the given timestep
        """
        return self.__data[idx]

    def ConvertArray(self, arr):
        """Converts the numpy array to a vtkDataArray
        """
        # Put raw data into vtk array
        data = nps.numpy_to_vtk(num_array=arr, deep=True, array_type=self.__vtktype)
        data.SetName(self.__dataName)
        return data


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to request data for current timestep
        """
        # Get output:
        output = vtk.vtkTable.GetData(outInfo)
        if self.NeedToRead():
            self._ReadUpFront()
        # Get requested time index
        i = _helpers.GetRequestedTime(self, outInfo)
        # Generate the data object
        arr = self._GetRawData(idx=i)
        data = self.ConvertArray(arr)
        output.AddColumn(data)
        return 1


    #### Seters and Geters ####


    def SetEndian(self, endian):
        """The endianness of the data file.

        Args:
            endian (int or char): no preference = '' or 0, Little = 1 or `<` or Big = 2 `>`.
        """
        pos = ['', '<', '>']
        if isinstance(endian, int):
            endian = pos[endian]
        if endian != self.__endian:
            self.__endian = endian
            self.__dtype, self.__vtktype = _helpers.getdTypes(dtype=self.__dtypechar, endian=self.__endian)
            self.Modified()

    def GetEndian(self):
        return self.__endian

    def SetDataType(self, dtype):
        """The data type of the binary file: `double='d'`, `float='f'`, `int='i'`
        """
        pos = ['d', 'f', 'i']
        if isinstance(dtype, int):
            dtype = pos[dtype]
        if dtype != self.__dtype:
            self.__dtypechar = dtype
            self.__dtype, self.__vtktype = _helpers.getdTypes(dtype=self.__dtypechar, endian=self.__endian)
            self.Modified()

    def GetDataTypes(self):
        return self.__dtype, self.__vtktype

    def SetDataName(self, dataName):
        """The string name of the data array generated from the inut file.
        """
        if dataName != self.__dataName:
            self.__dataName = dataName
            self.Modified(readAgain=False) # Don't re-read. Just request data again


    def GetDataName(self):
        return self.__dataName



class MadagascarReader(PackedBinariesReader):
    """This reads in float or double data that is packed into a Madagascar binary file format with a leader header. The reader ignores all of the ascii header details by searching for the sequence of three special characters: EOL EOL EOT and it will treat the followng binary packed data as one long array and make a ``vtkTable`` with one column of that data. The reader uses defaults to import as floats with native endianness. Use the Table to Uniform Grid or the Reshape Table filters to give more meaning to the data. We will later implement the ability to create a gridded volume from the header info. This reader is a quick fix for Samir. We chose to use a ``vtkTable`` object as the output of this reader because it gives us more flexibility in the filters we can apply to this data down the pipeline and keeps thing simple when using filters in this repository.
    `Details Here`_.

    .. _Details Here: http://www.ahay.org/wiki/RSF_Comprehensive_Description#Single-stream_RSF
    """
    __displayname__ = 'Madagascar SSRSF Reader'
    __type__ = 'reader'
    def __init__(self, **kwargs):
        PackedBinariesReader.__init__(self, **kwargs)

    def _ReadRawFile(self, fileName):
        dtype, vtktype = self.GetDataTypes()
        CTLSEQ = b'\014\014\004' # The control sequence to seperate header from data
        rpl = b''
        raw = []
        with open(fileName, 'rb') as file:
            raw = file.read()
            idx = raw.find(CTLSEQ)
            if idx == -1:
                warnings.warn('This is not a single stream RSF format file. Treating entire file as packed binary data.')
            else:
                raw = raw[idx:] # deletes the header
                raw = raw.replace(CTLSEQ, rpl) # removes the control sequence
        arr = np.fromstring(raw, dtype=dtype)
        return arr



__all__ = [
    'ManySlicesAlongPoints',
    'ManySlicesAlongAxis',
    'SliceThroughTime'
]

import vtk
from vtk.util import numpy_support as nps
import numpy as np
from vtk.numpy_interface import dataset_adapter as dsa
from datetime import datetime
# Import Helpers:
from ..base import FilterBase
from .. import _helpers
# NOTE: internal import - from scipy.spatial import cKDTree



class _SliceBase(FilterBase):
    """A helper class for making slicing fileters

    Note:
        * Make sure the input data source is slice-able.
        * The SciPy module is required for this filter.
    """
    __displayname__ = 'Base Slicing Filter'
    __type__ = 'filter'
    def __init__(self, numSlices=5,
            nInputPorts=1, inputType='vtkDataSet',
            nOutputPorts=1, outputType='vtkUnstructuredGrid'):
        FilterBase.__init__(self,
            nInputPorts=nInputPorts, inputType=inputType,
            nOutputPorts=nOutputPorts, outputType=outputType)
        # Parameters
        self.__numSlices = numSlices


    def _GeneratePlane(self, origin, normal):
        """Internal helper to build a ``vtkPlane`` for the cutter
        """
        # Get the slicing Plane:
        plane = vtk.vtkPlane() # Construct the plane object
        # Set the origin... needs to be inside of the grid
        plane.SetOrigin(origin[0], origin[1], origin[2])
        # set normal of that plane so we look at XZ section
        plane.SetNormal(normal)
        return plane

    def _Slice(self, pdi, pdo, plane):
        """Slice an input on a plane and produce the output
        """
        # create slice
        cutter = vtk.vtkCutter() # Construct the cutter object
        cutter.SetInputData(pdi) # Use the grid as the data we desire to cut
        cutter.SetCutFunction(plane) # the the cutter to use the plane we made
        cutter.Update() # Perfrom the Cut
        slc = cutter.GetOutput() # grab the output
        pdo.ShallowCopy(slc)

        return pdo

    def GetNumberOfSlices(self):
        return self.__numSlices

    def SetNumberOfSlices(self, num):
        if self.__numSlices != num:
            self.__numSlices = num
            self.Modified()


###############################################################################


class ManySlicesAlongPoints(_SliceBase):
    """Takes a series of points and a data source to be sliced. The points are used to construct a path through the data source and a slice is added at intervals of that path along the vector of that path at that point. This constructs many slices through the input dataset as an appended output ``vtkUnstructuredGrid``.

    Note:
        * Make sure the input data source is slice-able.
        * The SciPy module is required for this filter.
    """
    __displayname__ = 'Many Slices Along Points'
    __type__ = 'filter'
    def __init__(self, numSlices=5, nearestNbr=True):
        _SliceBase.__init__(self, numSlices=numSlices,
            nInputPorts=2, inputType='vtkDataSet',
            nOutputPorts=1, outputType='vtkUnstructuredGrid')
        self.__useNearestNbr = nearestNbr

    # CRITICAL for multiple input ports
    def FillInputPortInformation(self, port, info):
        """This simply makes sure the user selects the correct inputs
        """
        typ = 'vtkDataSet'
        if port == 0:
            typ = 'vtkPolyData' # Make sure points are poly data
        info.Set(self.INPUT_REQUIRED_DATA_TYPE(), typ)
        return 1

    def _ManySlicesAlongPoints(self, pdipts, pdidata, pdo):
        """Internal helper to perfrom the filter
        """
        # Get the Points over the NumPy interface
        wpdi = dsa.WrapDataObject(pdipts) # NumPy wrapped points
        points = np.array(wpdi.Points) # New NumPy array of points so we dont destroy input
        numPoints = pdipts.GetNumberOfPoints()
        if self.__useNearestNbr:
            from scipy.spatial import cKDTree # NOTE: Must have SciPy in ParaView
            tree = cKDTree(points)
            ptsi = tree.query(points[0], k=numPoints)[1]
        else:
            ptsi = [i for i in range(numPoints)]

        # iterate of points in order (skips last point):
        app = vtk.vtkAppendFilter()
        for i in range(0, numPoints - 1, numPoints/self.GetNumberOfSlices()):
            # get normal
            pts1 = points[ptsi[i]]
            pts2 = points[ptsi[i+1]]
            x1, y1, z1 = pts1[0], pts1[1], pts1[2]
            x2, y2, z2 = pts2[0], pts2[1], pts2[2]
            normal = [x2-x1,y2-y1,z2-z1]

            # create slice
            plane = self._GeneratePlane([x1,y1,z1], normal)
            temp = vtk.vtkPolyData()
            self._Slice(pdidata, temp, plane)
            app.AddInputData(temp)
        app.Update()
        pdo.ShallowCopy(app.GetOutput())
        return pdo



    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        # Get input/output of Proxy
        pdipts = self.GetInputData(inInfo, 0, 0) # Port 0: points
        pdidata = self.GetInputData(inInfo, 1, 0) # Port 1: sliceable data
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._ManySlicesAlongPoints(pdipts, pdidata, pdo)
        return 1



    #### Getters / Setters ####

    def SetUseNearestNbr(self, flag):
        """Set a flag on whether to use SciPy's nearest neighbor approximation when generating the slicing path
        """
        if self.__useNearestNbr != flag:
            self.__useNearestNbr = flag
            self.Modified()

    def Apply(self, points, data):
        self.SetInputDataObject(0, points)
        self.SetInputDataObject(1, data)
        self.Update()
        return self.GetOutput()


###############################################################################



class ManySlicesAlongAxis(_SliceBase):
    """Slices a ``vtkDataSet`` along a given axis many times.
    This produces a specified number of slices at once each with a normal vector
    oriented along the axis of choice and spaced uniformly through the range of
    the dataset on the chosen axis.
    """
    __displayname__ = 'Many Slices Along Axis'
    __type__ = 'filter'
    def __init__(self, numSlices=5, axis=0, rng=None, outputType='vtkUnstructuredGrid'):
        _SliceBase.__init__(self, numSlices=numSlices,
            nInputPorts=1, inputType='vtkDataSet',
            nOutputPorts=1, outputType=outputType)
        # Parameters
        self.__axis = axis
        self.__rng = rng


    def _GetOrigin(self, pdi, idx):
        """Internal helper to get plane origin
        """
        og = list(self.GetInputCenter(pdi))
        og[self.__axis] = self.__rng[idx]
        return og


    def GetInputBounds(self, pdi):
        """Gets the bounds of the input data set on the set slicing axis.
        """
        bounds = pdi.GetBounds()
        return bounds[self.__axis*2], bounds[self.__axis*2+1]

    def GetInputCenter(self, pdi):
        """Gets the center of the input data set

        Return:
            tuple: the XYZ coordinates of the center of the data set.
        """
        bounds = pdi.GetBounds()
        x = bounds[1] - bounds[0]
        y = bounds[3] - bounds[2]
        z = bounds[5] - bounds[4]
        return (x, y, z)

    def GetNormal(self):
        """Get the normal of the slicing plane"""
        norm = [0,0,0]
        norm[self.__axis] = 1
        return norm

    def _SetAxialRange(self, pdi):
        """Internal helper to set the slicing range along the set axis
        """
        bounds = self.GetInputBounds(pdi)
        self.__rng = np.linspace(bounds[0]+0.001, bounds[1]-0.001, num=self.GetNumberOfSlices())

    def _UpdateNumOutputs(self, num):
        """for internal use only
        """
        return 1



    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        self._SetAxialRange(pdi)
        normal = self.GetNormal()
        # Perfrom task
        app = vtk.vtkAppendFilter()
        for i in range(self.GetNumberOfSlices()):
            temp = vtk.vtkPolyData()
            origin = self._GetOrigin(pdi, i)
            plane = self._GeneratePlane(origin, normal)
            # Perfrom slice for that index
            self._Slice(pdi, temp, plane)
            app.AddInputData(temp)
        app.Update()
        pdo.ShallowCopy(app.GetOutput())

        return 1



    #### Getters / Setters ####


    def SetAxis(self, axis):
        """Set the axis on which to slice

        Args:
            axis (int): the axial index (0, 1, 2) = (x, y, z)
        """
        if axis not in (0,1,2):
            raise _helpers.PVGeoError('Axis choice must be 0, 1, or 2 (x, y, or z)')
        if self.__axis != axis:
            self.__axis = axis
            self.Modified()

    def GetRange(self):
        """Get the slicing range for the set axis
        """
        return self.__rng

    def GetAxis(self):
        """Get the set axis to slice upon as int index (0,1,2)
        """
        return self.__axis



###############################################################################


class SliceThroughTime(ManySlicesAlongAxis):
    """Takes a sliceable ``vtkDataSet`` and progresses a slice of it along a given axis. The macro requires that the clip already exist in the pipeline. This is especially useful if you have many clips linked together as all will move through the seen as a result of this macro.
    """
    __displayname__ = 'Slice Through Time'
    __type__ = 'filter'
    def __init__(self, numSlices=5, dt=1.0, axis=0, rng=None,):
        ManySlicesAlongAxis.__init__(self, numSlices=numSlices,
                axis=axis, rng=rng, outputType='vtkPolyData')
        # Parameters
        self.__dt = dt
        self.__timesteps = None


    def _UpdateTimeSteps(self):
        """For internal use only
        """
        self.__timesteps = _helpers.UpdateTimeSteps(self, self.GetNumberOfSlices(), self.__dt)



    #### Algorithm Methods ####

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        self._SetAxialRange(pdi)
        i = _helpers.GetRequestedTime(self, outInfo)
        # Perfrom task
        normal = self.GetNormal()
        origin = self._GetOrigin(pdi, i)
        plane = self._GeneratePlane(origin, normal)
        self._Slice(pdi, pdo, plane)
        return 1

    def RequestInformation(self, request, inInfoVec, outInfoVec):
        """Used by pipeline to set the time information
        """
        # register time:
        self._UpdateTimeSteps()
        return 1

    #### Public Getters / Setters ####

    def SetNumberOfSlices(self, num):
        """Set the number of slices/timesteps to generate
        """
        ManySlicesAlongAxis.SetNumberOfSlices(self, num)
        self._UpdateTimeSteps()
        self.Modified()

    def SetTimeDelta(self, dt):
        """
        Set the time step interval in seconds
        """
        if self.__dt != dt:
            self.__dt = dt
            self._UpdateTimeSteps()
            self.Modified()

    def GetTimestepValues(self):
        """Use this in ParaView decorator to register timesteps
        """
        return self.__timesteps.tolist() if self.__timesteps is not None else None



__all__ = [
    'checkNumpy',
]
__displayname__ = 'Version Verifier'

def checkNumpy(warn=True):
    """A method to check the active environments version of NumPy for
    compatibility with PVGeo.

    Args:
        war (bool): raise a RuntimeWarning if NumPy is not at satisfactory version.
    """
    import numpy as np
    v = np.array(np.__version__.split('.')[0:2], dtype=int)
    if v[0] >= 1 and v[1] >= 10:
        return True
    elif warn:
        raise RuntimeWarning('WARNING: Your version of NumPy is below 1.10.x (you are using %s), please update the NumPy module used in ParaView for performance enhancement. Some filters/readers may crash otherwise.' % np.__version__)
    return False

#needToUpdateNumPy = checkNumpy()



__all__ = [
    'DelimitedTextReader',
    'XYZTextReader'
]

import numpy as np

# Import Helpers:
from ..base import ReaderBase
from .. import _helpers


class DelimitedTextReader(ReaderBase):
    """This reader will take in any delimited text file and make a ``vtkTable`` from it. This is not much different than the default .txt or .csv reader in ParaView, however it gives us room to use our own extensions and a little more flexibility in the structure of the files we import.
    """
    __displayname__ = 'Delimited Text Reader'
    __type__ = 'reader'
    def __init__(self, nOutputPorts=1, outputType='vtkTable', **kwargs):
        ReaderBase.__init__(self,
            nOutputPorts=nOutputPorts, outputType=outputType,
            **kwargs)

        # Parameters to control the file read:
        #- if these are set/changed, we must reperform the read
        self.__delimiter = kwargs.get('delimiter', ' ')
        self.__useTab = kwargs.get('useTab', False)
        self.__skipRows = kwargs.get('skiprows', 0)
        self.__comments = kwargs.get('comments', '!')
        self.__hasTitles = kwargs.get('HasTitles', True)
        # Data objects to hold the read data for access by the pipeline methods
        self.__data = []
        self.__titles = []

    def _GetDeli(self):
        """For itenral use
        """
        if self.__useTab:
            return None
        return self.__delimiter

    #### Methods for performing the read ####

    def _GetFileContents(self, idx=None):
        if idx is not None:
            fileNames = [self.GetFileNames(idx=idx)]
        else:
            fileNames = self.GetFileNames()
        contents = []
        for f in fileNames:
            try:
                contents.append(np.genfromtxt(f, dtype=str, delimiter='\n', comments=self.__comments)[self.__skipRows::])
            except (FileNotFoundError, OSError) as fe:
                raise _helpers.PVGeoError(str(fe))
        if idx is not None: return contents[0]
        return contents

    def _ExtractHeader(self, content):
        """Override this. Removes header from single file's content.
        """
        if len(np.shape(content)) > 2:
            raise _helpers.PVGeoError("`_ExtractHeader()` can only handle a sigle file's content")
        idx = 0
        if self.__hasTitles:
            titles = content[idx].split(self._GetDeli())
            idx += 1
        else:
            cols = len(content[idx].split(self._GetDeli()))
            titles = []
            for i in range(cols):
                titles.append('Field %d' % i)
        return titles, content[idx::]

    def __ExtractHeaders(self, contents):
        """Should NOT be overriden.
        """
        ts = []
        for i in range(len(contents)):
            titles, newcontent = self._ExtractHeader(contents[i])
            contents[i] = newcontent
            ts.append(titles)
        # Check that the titles are the same across files:
        ts = np.unique(np.asarray(ts), axis=0)
        if len(ts) > 1:
            raise _helpers.PVGeoError('Data array titles varied across file timesteps. This data is invalid as a timeseries.')
        return ts[0], contents


    def __FileContentsToDataArray(self, contents):
        """Should NOT need to be overriden
        """
        data = []
        for content in contents:
            data.append(np.genfromtxt((line.encode('utf8') for line in content), delimiter=self._GetDeli(), dtype=None))
        return data

    def _ReadUpFront(self):
        """Should not need to be overridden.
        """
        # Perform Read
        contents = self._GetFileContents()
        self.__titles, contents = self.__ExtractHeaders(contents)
        self.__data = self.__FileContentsToDataArray(contents)
        self.NeedToRead(flag=False)
        return 1

    #### Methods for accessing the data read in #####

    def _GetRawData(self, idx=0):
        """This will return the proper data for the given timestep.
        """
        return self.__data[idx]


    #### Algorithm Methods ####

    def RequestData(self, request, inInfo, outInfo):
        """Used b pipeline to get data for current timestep and populate the output data object.
        """
        # Get output:
        output = self.GetOutputData(outInfo, 0)
        # Get requested time index
        i = _helpers.GetRequestedTime(self, outInfo)
        if self.NeedToRead():
            self._ReadUpFront()
        # Generate the data object
        _helpers.placeArrInTable(self._GetRawData(idx=i), self.__titles, output)
        return 1


    #### Seters and Geters ####


    def SetDelimiter(self, deli):
        """The input file's delimiter. To use a tab delimiter please use ``SetUseTab()``

        Args:
            deli (str): a string delimiter/seperator
        """
        if deli != self.__delimiter:
            self.__delimiter = deli
            self.Modified()

    def SetSplitOnWhiteSpace(self, flag):
        """Set a boolean flag to override the ``SetDelimiter()`` and use any white space as a delimiter.
        """
        if flag != self.__useTab:
            self.__useTab = flag
            self.Modified()

    def SetUseTab(self, flag):
        """Deprecated"""
        self.SetSplitOnWhiteSpace(flag)


    def SetSkipRows(self, skip):
        """The integer number of rows to skip at the top of the file.
        """
        if skip != self.__skipRows:
            self.__skipRows = skip
            self.Modified()

    def GetSkipRows(self):
        return self.__skipRows

    def SetComments(self, identifier):
        """The character identifier for comments within the file.
        """
        if identifier != self.__comments:
            self.__comments = identifier
            self.Modified()

    def SetHasTitles(self, flag):
        """A boolean for if the delimited file has header titles for the data arrays.
        """
        if self.__hasTitles != flag:
            self.__hasTitles = flag
            self.Modified()

    def HasTitles(self):
        return self.__hasTitles

    def GetTitles(self):
        return self.__titles




class XYZTextReader(DelimitedTextReader):
    """A makeshift reader for XYZ files where titles have comma delimiter and data has space delimiter.
    """
    __displayname__ = 'XYZ Text Reader'
    __type__ = 'reader'
    def __init__(self, **kwargs):
        DelimitedTextReader.__init__(self, **kwargs)
        self.SetComments(kwargs.get('comments', '#'))

    # Simply override the extract titles functionality
    def _ExtractHeader(self, content):
        titles = content[0][2::].split(', ') # first two characers of header is '! '
        return titles, content[1::]



import os
import sys
import platform

def GetInstallationPaths():
    path = os.path.dirname(os.path.dirname(__file__))
    PYTHONPATH = path
    PV_PLUGIN_PATH = '%s/%s' % (path, 'PVPlugins')

    script = 'curl -s  https://cdn.rawgit.com/OpenGeoVis/PVGeo/be9e9a95/installMac.sh | sh -s'

    if 'darwin' in platform.system().lower():
        # Install launch agents
        print('Copy paste the following line(s) to execute in your bash terminal:\n')
        print('%s %s' % (script, PYTHONPATH))
        print('\n')
    else:
        print('Set these environmental variables to use PVGeo in ParaView:')
        print('\n')
        print('export PYTHONPATH="%s"' % PYTHONPATH)
        print('export PV_PLUGIN_PATH="%s"' % PV_PLUGIN_PATH)
        print('\n')
    return




if __name__ == '__main__':
    from .tester import test
    arg = sys.argv[1]
    if arg.lower() == 'test':
        test(True)
    elif arg.lower() == 'install':
        GetInstallationPaths()
    else:
        raise RuntimeError('Unknown argument: %s' % arg)



__all__ = [
    'numToVTK',
    'getSelectedArrayName',
    'getSelectedArrayField',
    'copyArraysToPointData',
    'getArray',
    'addArray',
    'getSelectedArray',
    #'GetDecimalPlaces',
]

import vtk
import numpy as np
from vtk.util import numpy_support as nps
from . import errors as _helpers

def numToVTK(arr, name):
    """Converts a 1D numpy array to a VTK data array given a nameself.

    Args:
        arr (np.array) : A 1D numpy array
        name (str): the name of the data array for VTK

    Return:
        vtkDataArray : a converted data array
    """
    c = nps.numpy_to_vtk(num_array=arr, deep=True)
    c.SetName(name)
    return c


def getSelectedArrayName(algorithm, idx):
    """Gets the name of the input array for a given index on a VTK algorithm

    Args:
        algorithm (vtkAlgorithm): A vtkAlgorithm class instantiation
        idx (int): the input array index

    Return:
        str : the name of the input array for the given index
    """
    info = algorithm.GetInputArrayInformation(idx)
    return info.Get(vtk.vtkDataObject.FIELD_NAME())


def getSelectedArrayField(algorithm, idx):
    """Gets the field of the input array for a given index on a VTK algorithm

    Args:
        algorithm (vtkAlgorithm) : A vtkAlgorithm class instantiation
        idx (int) : the input array index

    Return:
        int : the field type of the input array for the given index
    """
    info = algorithm.GetInputArrayInformation(idx)
    return info.Get(vtk.vtkDataObject.FIELD_ASSOCIATION())


def copyArraysToPointData(pdi, pdo, field):
    """Copys arrays from an input to an ouput's point data.

    Args:
        pdi (vtkDataObject) : The input data object to copy from
        pdo (vtkDataObject) : The output data object to copy over to
        field (int) : the field type id

    Return:
        vtkDataObject : returns the output data object parameter
    """
    # Point Data
    if field == 0:
        for i in range(pdi.GetPointData().GetNumberOfArrays()):
            arr = pdi.GetPointData().GetArray(i)
            pdo.GetPointData().AddArray(arr)
    # Cell Data: DO NOT USE
    elif field == 1:
        for i in range(pdi.GetCellData().GetNumberOfArrays()):
            arr = pdi.GetCellData().GetArray(i)
            pdo.GetPointData().AddArray(arr)
    # Field Data:
    elif field == 2:
        for i in range(pdi.GetFieldData().GetNumberOfArrays()):
            arr = pdi.GetFieldData().GetArray(i)
            pdo.GetPointData().AddArray(arr)
    # Row Data:
    elif field == 6:
        for i in range(pdi.GetRowData().GetNumberOfArrays()):
            arr = pdi.GetRowData().GetArray(i)
            pdo.GetPointData().AddArray(arr)
    else:
        raise _helpers.PVGeoError('Field association not defined. Try inputing Point, Cell, Field, or Row data.')

    # Field Data
    return pdo


def getArray(wpdi, field, name):
    """Grabs an array from vtkDataObject given its name and field association.

    Args:
        wpdi  (wrapped vtkDataObject) : the input data object wrapped using vtk dataset adapter
        field (int) : the field type id
        name (str) : the name of the input array for the given index

    Return:
        numpy.array : a wrapped ``vtkDataArray`` for NumPy
    """
    # Point Data
    if field == 0:
        arr = wpdi.PointData[name]
    # Cell Data:
    elif field == 1:
        arr = wpdi.CellData[name]
    # Field Data:
    elif field == 2:
        arr = wpdi.FieldData[name]
    # Row Data:
    elif field == 6:
        arr = wpdi.RowData[name]
    else:
        raise _helpers.PVGeoError('Field association not defined. Try inputing Point, Cell, Field, or Row data.')
    return arr


def getSelectedArray(algorithm, wpdi, idx):
    """Gets selectected array at index idx wrapped for NumPy

    Args:
        algorithm (vtkAlgorithm) : A vtkAlgorithm class instantiation
        wpdi (wrapped vtkDataObject) : the input data object wrapped using vtk dataset adapter
        idx (int) : the input array index

    Return:
        numpy.array : a wrapped ``vtkDataArray`` for NumPy
    """
    name = getSelectedArrayName(algorithm, idx)
    field = getSelectedArrayField(algorithm, idx)
    return getArray(wpdi, field, name)

def addArray(pdo, field, vtkArray):
    """Adds an array to a vtkDataObject given its field association.

    Args:
        pdo (vtkDataObject) : the output data object
        field (int) : the field type id
        vtkArray (vtkDataArray) : the data array to add to the output

    Return:
        vtkDataObject : the output data object with the data array added
    """
    # Point Data
    if field == 0:
        pdo.GetPointData().AddArray(vtkArray)
    # Cell Data:
    elif field == 1:
        pdo.GetCellData().AddArray(vtkArray)
    # Field Data:
    elif field == 2:
        pdo.GetFieldData().AddArray(vtkArray)
    # Row Data:
    elif field == 6:
        pdo.GetRowData().AddArray(vtkArray)
    else:
        raise _helpers.PVGeoError('Field association not defined. Try inputing Point, Cell, Field, or Row data.')
    return pdo


# def GetDecimalPlaces(arr, barrier=6):
#     arr = np.array(arr.flatten(), dtype=str)
#     num = 0
#     for val in arr:
#         n = len(str(val).split('.')[1])
#         if n > num:
#             num = n
#     # Now do not let exceed barrier
#     if num > barrier: return barrier
#     return num



__all__ = [
    'VoxelizePoints',
]

import numpy as np
import vtk
from vtk.util import keys
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa

from ..base import FilterBase
from .. import _helpers
from ..version import checkNumpy
from .xyz import RotationTool

###############################################################################


class VoxelizePoints(FilterBase):
    """This makes a ``vtkUnstructuredGrid`` of scattered points given voxel sizes as input arrays.
    This assumes that the data is at least 2-Dimensional on the XY Plane.
    """
    __displayname__ = 'Voxelize Points'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkPolyData',
            nOutputPorts=1, outputType='vtkUnstructuredGrid')
        self.__dx = kwargs.get('dx', None)
        self.__dy = kwargs.get('dy', None)
        self.__dz = kwargs.get('dz', None)
        self.__estimateGrid = kwargs.get('estimate', True)
        self.__safe = kwargs.get('safe', 10.0)

        # Not controlled by user:
        self.__angle = 0.0



    def AddFieldData(self, grid):
        """An internal helper to add the recovered information as field data
        """
        # Add angle
        a = vtk.vtkDoubleArray()
        a.SetName('Recovered Angle (Deg.)')
        a.SetNumberOfValues(1)
        a.SetValue(0, np.rad2deg(self.__angle))
        grid.GetFieldData().AddArray(a)
        # Add cell sizes
        s = vtk.vtkDoubleArray()
        s.SetName('Recovered Cell Sizes')
        s.SetNumberOfComponents(3)
        s.InsertNextTuple3(self.__dx, self.__dy, self.__dz)
        grid.GetFieldData().AddArray(s)
        return grid

    @staticmethod
    def AddCellData(grid, arr, name):
        """Add a NumPy array as cell data to the given grid input
        """
        c = nps.numpy_to_vtk(num_array=arr, deep=True)
        c.SetName(name)
        grid.GetCellData().AddArray(c)
        return grid


    def EstimateUniformSpacing(self, x, y, z):
        """This assumes that the input points make up some sort of uniformly spaced
        grid on at least an XY Plane
        """
        # TODO: implement ability to rotate around Z axis (think PoroTomo vs UTM)
        # TODO: implement way to estimate rotation
        assert(len(x) == len(y) == len(z))
        num = len(x)
        if num == 1:
            # Only one point.. use safe
            return x, y, z, self.__safe, self.__safe, self.__safe, 0.0

        r = RotationTool()
        xr, yr, zr, dx, dy, angle = r.EstimateAndRotate(x, y, z)
        self.__angle = angle
        uz = np.diff(np.unique(z))
        if len(uz) > 0: dz = np.average(uz)
        else: dz = self.__safe
        self.__dx = dx
        self.__dy = dy
        self.__dz = dz
        return xr, yr, zr


    def PointsToGrid(self, xo,yo,zo, dx,dy,dz, grid=None):
        """Convert XYZ points to a ``vtkUnstructuredGrid``.
        """
        if not checkNumpy():
            raise _helpers.PVGeoError("`VoxelizePoints` cannot work with versions of NumPy below 1.10.x . You must update NumPy.")
            return None
        if grid is None:
            grid = vtk.vtkUnstructuredGrid()

        # TODO: Check dtypes on all arrays. Need to be floats

        if self.__estimateGrid:
            x,y,z = self.EstimateUniformSpacing(xo, yo, zo)
        else:
            x,y,z = xo, yo, zo

        dx,dy,dz = self.__dx, self.__dy, self.__dz
        if isinstance(dx, np.ndarray) and len(dx) != len(x):
            raise _helpers.PVGeoError('X-Cell spacings are not properly defined for all points.')
        if isinstance(dy, np.ndarray) and len(dy) != len(y):
            raise _helpers.PVGeoError('X-Cell spacings are not properly defined for all points.')
        if isinstance(dz, np.ndarray) and len(dz) != len(z):
            raise _helpers.PVGeoError('X-Cell spacings are not properly defined for all points.')

        numCells = len(x)

        # Generate cell nodes for all points in data set
        #- Bottom
        c_n1 = np.stack( ((x - dx/2) , (y - dy/2), (z - dz/2) ), axis=1)
        c_n2 = np.stack(( (x + dx/2) , (y - dy/2), (z - dz/2) ), axis=1)
        c_n3 = np.stack(( (x - dx/2) , (y + dy/2), (z - dz/2) ), axis=1)
        c_n4 = np.stack(( (x + dx/2) , (y + dy/2), (z - dz/2) ), axis=1)
        #- Top
        c_n5 = np.stack(( (x - dx/2) , (y - dy/2), (z + dz/2) ), axis=1)
        c_n6 = np.stack(( (x + dx/2) , (y - dy/2), (z + dz/2) ), axis=1)
        c_n7 = np.stack(( (x - dx/2) , (y + dy/2), (z + dz/2) ), axis=1)
        c_n8 = np.stack(( (x + dx/2) , (y + dy/2), (z + dz/2) ), axis=1)

        #- Concatenate
        all_nodes = np.concatenate((
            c_n1,
            c_n2,
            c_n3,
            c_n4,
            c_n5,
            c_n6,
            c_n7,
            c_n8), axis=0)

        # Search for unique nodes and use the min cell size as the tolerance
        TOLERANCE = np.min([dx, dy]) / 2.0
        # Round XY plane by the tolerance
        txy = np.around(all_nodes[:,0:2]/TOLERANCE)
        all_nodes[:,0:2] = txy
        unique_nodes, ind_nodes = np.unique(all_nodes, return_inverse=True, axis=0)
        unique_nodes[:,0:2] *= TOLERANCE
        numPts = len(unique_nodes)

        # Make the cells
        pts = vtk.vtkPoints()
        cells = vtk.vtkCellArray()

        # insert unique nodes as points
        if self.__estimateGrid:
            unique_nodes[:,0:2] = RotationTool.Rotate(unique_nodes[:,0:2], -self.__angle)
            self.AddFieldData(grid)

        # Add unique nodes as points in output
        pts.SetData(nps.numpy_to_vtk(unique_nodes))

        cnt = 0
        arridx = np.zeros(numCells)
        for i in range(numCells):
            # OPTIMIZE: may be complicated but can be speed up
            vox = vtk.vtkVoxel()
            for j in range(8):
                vox.GetPointIds().SetId(j, ind_nodes[j*numCells + i])
            cells.InsertNextCell(vox)

            arridx[i] = i
            cnt += 8

        grid.SetPoints(pts)
        grid.SetCells(vtk.VTK_VOXEL, cells)
        #VoxelizePoints.AddCellData(grid, arridx, 'Voxel ID') # For testing
        return grid

    def _CopyArrays(self, pdi, pdo):
        """internal helper to copy arrays from point data to cell data in the voxels.
        """
        for i in range(pdi.GetPointData().GetNumberOfArrays()):
            arr = pdi.GetPointData().GetArray(i)
            _helpers.addArray(pdo, 1, arr) # adds to CELL data
        return pdo

    def RequestData(self, request, inInfoVec, outInfoVec):
        """Used by pipeline to generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfoVec, 0, 0)
        pdo = self.GetOutputData(outInfoVec, 0)
        # Perfrom task
        wpdi = dsa.WrapDataObject(pdi)
        pts = wpdi.Points
        x, y, z = pts[:,0], pts[:,1], pts[:,2]
        self.PointsToGrid(x, y, z,
            self.__dx, self.__dy, self.__dz, grid=pdo)
        # Now append data to grid
        self._CopyArrays(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def SetSafeSize(self, safe):
        """A voxel size to use if a spacing cannot be determined for an axis
        """
        if self.__safe != safe:
            self.__safe = safe
            self.Modified()

    def SetDeltaX(self, dx):
        """Set the X cells spacing

        Args:
            dx (float or np.array(floats)): the spacing(s) for the cells in the X-direction
        """
        self.__dx = dx
        self.Modified()

    def SetDeltaY(self, dy):
        """Set the Y cells spacing

        Args:
            dy (float or np.array(floats)): the spacing(s) for the cells in the Y-direction
        """
        self.__dy = dy
        self.Modified()

    def SetDeltaZ(self, dz):
        """Set the Z cells spacing

        Args:
            dz (float or np.array(floats)): the spacing(s) for the cells in the Z-direction
        """
        self.__dz = dz
        self.SetSafeSize(np.min(dz))
        self.Modified()

    def SetDeltas(self, dx, dy, dz):
        """Set the cell spacings for each axial direction

        Args:
            dx (float or np.array(floats)): the spacing(s) for the cells in the X-direction
            dy (float or np.array(floats)): the spacing(s) for the cells in the Y-direction
            dz (float or np.array(floats)): the spacing(s) for the cells in the Z-direction
        """
        self.SetDeltaX(dx)
        self.SetDeltaY(dy)
        self.SetDeltaZ(dz)

    def SetEstimateGrid(self, flag):
        """Set a flag on whether or not to estimate the grid spacing/rotation
        """
        if self.__estimateGrid != flag:
            self.__estimateGrid = flag
            self.Modified()





###############################################################################



__all__ = [
    # OcTree
    'OcTreeReader',
    'OcTreeAppender',
]

import numpy as np
from vtk.util import numpy_support as nps
import vtk
import os

from ..base import AlgorithmBase
from .two_file_base import ubcMeshReaderBase, ModelAppenderBase
from .. import _helpers



class OcTreeReader(ubcMeshReaderBase):
    """This class reads a UBC OcTree Mesh file and builds a ``vtkUnstructuredGrid`` of the data in the file.
    Model File is optional. Reader will still construct ``vtkUnstructuredGrid`` safely.
    """
    __displayname__ = 'UBC OcTree Mesh Reader'
    __type__ = 'reader'
    def __init__(self, nOutputPorts=1, outputType='vtkUnstructuredGrid', **kwargs):
        ubcMeshReaderBase.__init__(self,
            nOutputPorts=nOutputPorts, outputType=outputType,
            **kwargs)

        self.__mesh = vtk.vtkUnstructuredGrid()
        self.__models = []


    @staticmethod
    def ubcOcTreeMesh(FileName, pdo=None):
        """This method reads a UBC OcTree Mesh file and builds a ``vtkUnstructuredGrid`` of the data in the file. This method generates the ``vtkUnstructuredGrid`` without any data attributes.

        Args:
            FileName (str): The mesh filename as an absolute path for the input mesh file in UBC OcTree format.
            pdo (vtkUnstructuredGrid): A pointer to the output data object.

        Returns:
            vtkUnstructuredGrid: Returns a ``vtkUnstructuredGrid`` generated from the UBCMesh grid. Mesh is defined by the input mesh file. No data attributes here, simply an empty mesh. Use the ``PlaceModelOnOcTreeMesh()`` method to associate with model data.
        """
        if pdo is None:
            pdo = vtk.vtkUnstructuredGrid() # vtkUnstructuredGrid

        #--- Read in the mesh ---#
        fileLines = np.genfromtxt(FileName, dtype=str,
            delimiter='\n', comments='!')

        # Get mesh dimensions
        dim = np.array(fileLines[0].
            split('!')[0].split(), dtype=int)
        # First three values are the number of cells in the core mesh and remaining 6 values are padding for the core region.
        pad = dim[3:6] # TODO: check if there because optional... might throw error if not there
        dim = dim[0:3]
        ne,nn,nz = dim[0], dim[1], dim[2]

        # The origin corner (Southwest-top)
        #- Remember UBC format specifies down as the positive Z
        #- Easting, Northing, Altitude
        oo = np.array(
            fileLines[1].split('!')[0].split(),
            dtype=float
        )
        oe,on,oz = oo[0],oo[1],oo[2]

        # Widths of the core cells in the Easting, Northing, and Vertical directions.
        ww = np.array(
            fileLines[2].split('!')[0].split(),
            dtype=float
        )
        we,wn,wz = ww[0],ww[1],ww[2]

        # Number of cells in OcTree mesh
        numCells = np.array(
            fileLines[3].split('!')[0].split(),
            dtype=float
        )

        # Read the remainder of the file containing the index arrays
        indArr = np.genfromtxt(
            (line.encode('utf8') for line in fileLines[4::]), dtype=np.int)

        # Start processing the information
        # Make vectors of the base mesh node, starting in the wsb corner
        vec_full_nx = np.cumsum(np.hstack((oe, we * np.ones(ne))))
        vec_full_ny = np.cumsum(np.hstack((on, wn * np.ones(nn))))
        vec_full_nz = np.cumsum(np.hstack((oz - wz * nz, wz * np.ones(nz))))
        # Make indices
        indC = indArr[:, 0:3] + np.array([-1, -1, -1])  # Shift to be 0 indexed
        # Flip the z-ind to start from bottom
        indC[:, 2] = nz - indC[:, 2] - indArr[:, 3]
        cell_size = np.reshape(indArr[:, 3], (len(indArr[:, 3]), 1))
        cell_zero = np.zeros((len(cell_size), 1), dtype=np.int)
        # Need to reference the nodal numbers to form the cell.
        # Find the 8 corners of each cell
        #
        #             z+   y+
        #             |  /
        #             | /
        #             |/_ _ _ x+
        #
        #    N7--------N8
        #   /|         /|
        #  N5--------N6 |
        #  | |        | |
        #  | N3-------|N4
        #  |/         |/
        #  N1--------N2

        # UBC Octree indexes always the top-left-close corner first
        # For to define the cells in UBC order
        cell_n1 = indC + np.hstack((cell_zero, cell_zero, cell_zero))  # Node 1 in all cells
        cell_n2 = indC + np.hstack((cell_size, cell_zero, cell_zero))  # Node 2 in all cells
        cell_n3 = indC + np.hstack((cell_zero, cell_size, cell_zero))  # Node 3 in all cells
        cell_n4 = indC + np.hstack((cell_size, cell_size, cell_zero))  # Node 4 in all cells
        cell_n5 = indC + np.hstack((cell_zero, cell_zero, cell_size))  # Node 5 in all cells
        cell_n6 = indC + np.hstack((cell_size, cell_zero, cell_size))  # Node 6 in all cells
        cell_n7 = indC + np.hstack((cell_zero, cell_size, cell_size))  # Node 7 in all cells
        cell_n8 = indC + np.hstack((cell_size, cell_size, cell_size))  # Node 8 in all cells
        # Sort the nodal index to be from the south-west-bottom most corner,
        # comply with SimPEG ordering
        # NOTE: Is not needed but prefered
        ind_cell_corner = np.argsort(
            cell_n1.view(','.join(3 * ['int'])), axis=0, order=('f2', 'f1', 'f0'))
        sortcell_n1 = cell_n1[ind_cell_corner][:, 0, :]
        sortcell_n2 = cell_n2[ind_cell_corner][:, 0, :]
        sortcell_n3 = cell_n3[ind_cell_corner][:, 0, :]
        sortcell_n4 = cell_n4[ind_cell_corner][:, 0, :]
        sortcell_n5 = cell_n5[ind_cell_corner][:, 0, :]
        sortcell_n6 = cell_n6[ind_cell_corner][:, 0, :]
        sortcell_n7 = cell_n7[ind_cell_corner][:, 0, :]
        sortcell_n8 = cell_n8[ind_cell_corner][:, 0, :]
        # Find the unique nodes
        all_nodes = np.concatenate((
            sortcell_n1,
            sortcell_n2,
            sortcell_n3,
            sortcell_n4,
            sortcell_n5,
            sortcell_n6,
            sortcell_n7,
            sortcell_n8), axis=0)
        # Make a rec array to search for uniques
        all_nodes_rec = all_nodes.view(','.join(3 * ['int']))[:, 0]
        unique_nodes, ind_nodes_vec = np.unique(all_nodes_rec, return_inverse=True)

        # Reshape the matrix
        ind_nodes_mat = ind_nodes_vec.reshape(((8, int(ind_nodes_vec.size / 8)))).T
        ind_nodes_full = np.concatenate((
            np.ones((
                ind_nodes_mat.shape[0],
                1), dtype=np.int64) * ind_nodes_mat.shape[1],
            ind_nodes_mat), axis=1).ravel()

        # Make the VTK object
        # Make the points.
        ptsArr = np.concatenate((
            vec_full_nx[unique_nodes['f0']].reshape(-1, 1),
            vec_full_ny[unique_nodes['f1']].reshape(-1, 1),
            vec_full_nz[unique_nodes['f2']].reshape(-1, 1)), axis=1)
        vtkPtsData = nps.numpy_to_vtk(ptsArr, deep=1)
        vtkPts = vtk.vtkPoints()
        vtkPts.SetData(vtkPtsData)

        # Make the cells
        # Cells -cell array
        CellArr = vtk.vtkCellArray()
        CellArr.SetNumberOfCells(numCells)
        CellArr.SetCells(
            numCells,
            nps.numpy_to_vtkIdTypeArray(
                np.ascontiguousarray(ind_nodes_full), deep=1))

        # Construct the VTK object declared in `pdo`
        # Set the objects properties
        pdo.SetPoints(vtkPts)
        pdo.SetCells(vtk.VTK_VOXEL, CellArr)

        # Add the indexing of the cell's
        vtkIndexArr = nps.numpy_to_vtk(
            np.ascontiguousarray(ind_cell_corner.ravel()), deep=1)
        vtkIndexArr.SetName('index_cell_corner')
        pdo.GetCellData().AddArray(vtkIndexArr)

        return pdo

    @staticmethod
    def PlaceModelOnOcTreeMesh(mesh, model, dataNm='Data'):
        """Places model data onto a mesh. This is for the UBC Grid data reaers to associate model data with the mesh grid.

        Args:
            mesh (vtkUnstructuredGrid): The ``vtkUnstructuredGrid`` that is the mesh to place the model data upon. Needs to have been read in by ubcOcTree
            model (np.ndarray): A NumPy float array that holds all of the data to place inside of the mesh's cells.
            dataNm (str): The name of the model data array once placed on the ``vtkUnstructuredGrid``.

        Returns:
            vtkUnstructuredGrid: Returns the input ``vtkUnstructuredGrid`` with model data appended.
        """
        if type(model) is dict:
            for key in model.keys():
                mesh = OcTreeReader.PlaceModelOnOcTreeMesh(mesh, model[key], dataNm=key)
            return mesh
        # Make sure this model file fits the dimensions of the mesh
        numCells = mesh.GetNumberOfCells()
        if (numCells < len(model)):
            raise _helpers.PVGeoError('This model file has more data than the given mesh has cells to hold.')
        elif (numCells > len(model)):
            raise _helpers.PVGeoError('This model file does not have enough data to fill the given mesh\'s cells.')

        # This is absolutely crucial!
        # Do not play with unless you know what you are doing!
        ind_reorder = nps.vtk_to_numpy(
            mesh.GetCellData().GetArray('index_cell_corner'))

        model = model[ind_reorder]

        # Convert data to VTK data structure and append to output
        c = nps.numpy_to_vtk(num_array=model, deep=True)
        c.SetName(dataNm)
        # THIS IS CELL DATA! Add the model data to CELL data:
        mesh.GetCellData().AddArray(c)
        return mesh



    def __ubcOcTree(self, FileName_Mesh, FileName_Models, output):
        """Wrapper to Read UBC GIF OcTree mesh and model file pairs. UBC OcTree models are defined using a 2-file format. The "mesh" file describes how the data is descritized. The "model" file lists the physical property values for all cells in a mesh. A model file is meaningless without an associated mesh file. This only handles OcTree formats

        Args:
            FileName_Mesh (str): The OcTree Mesh filename as an absolute path for the input mesh file in UBC OcTree Mesh Format
        FileName_Models (list(str)): The model filenames as absolute paths for the input model timesteps in UBC OcTree Model Format.
            output (vtkUnstructuredGrid): The output data object

        Returns:
            vtkUnstructuredGrid: Returns a ``vtkUnstructuredGrid`` generated from the UBC 2D/3D Mesh grid. Mesh is defined by the input mesh file. Cell data is defined by the input model file.
        """
        if self.NeedToReadMesh():
            # Construct/read the mesh
            OcTreeReader.ubcOcTreeMesh(FileName_Mesh, pdo=self.__mesh)
            self.NeedToReadMesh(flag=False)
        output.DeepCopy(self.__mesh)
        if self.NeedToReadModels() and self.ThisHasModels():
            # Read the model data
            self.__models = []
            for f in FileName_Models:
                # Read the model data
                self.__models.append(ubcMeshReaderBase.ubcModel3D(f))
            self.NeedToReadModels(flag=False)
        return output


    def RequestData(self, request, inInfo, outInfo):
        # Get output:
        output = self.GetOutputData(outInfo, 0)
        # Get requested time index
        i = _helpers.GetRequestedTime(self, outInfo)
        self.__ubcOcTree(
            self.GetMeshFileName(),
            self.GetModelFileNames(),
            output)

        # Place the model data for given timestep onto the mesh
        if len(self.__models) > i:
            OcTreeReader.PlaceModelOnOcTreeMesh(output, self.__models[i], self.GetDataName())

        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        """Pipeline method for handling requests about the grid extents and time
        step values
        """
        # Call parent to handle time stuff
        ubcMeshReaderBase.RequestInformation(self, request, inInfo, outInfo)
        # Now set whole output extent
        if self.NeedToReadMesh():
            ext = self._ReadExtent()
            info = outInfo.GetInformationObject(0)
            # Set WHOLE_EXTENT: This is absolutely necessary
            info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        return 1

    def ClearMesh(self):
        """Use to clean/rebuild the mesh.
        """
        self.__mesh = vtk.vtkUnstructuredGrid()
        ubcMeshReaderBase.ClearModels(self)

    def ClearModels(self):
        """Use to clean the models and reread the data
        """
        self.__models = []
        ubcMeshReaderBase.ClearModels(self)





################################################################################

class OcTreeAppender(ModelAppenderBase):
    """This filter reads a timeseries of models and appends it to an input
    ``vtkUnstructuredGrid``
    """
    __displayname__ = 'UBC OcTree Mesh Appender'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        ModelAppenderBase.__init__(self,
            inputType='vtkUnstructuredGrid',
            outputType='vtkUnstructuredGrid',
            **kwargs)


    def _ReadUpFront(self):
        reader = ubcMeshReaderBase.ubcModel3D
        self._models = []
        for f in self._modelFileNames:
            # Read the model data
            self._models.append(reader(f))
        self.NeedToRead(flag=False)
        return

    def _PlaceOnMesh(self, output, idx=0):
        OcTreeReader.PlaceModelOnOcTreeMesh(output, self._models[idx], self._dataname)
        return



__all__ = [
    'PointsToPolyData',
    'latLonTableToCartesian',
    'RotatePoints',
    'ExtractPoints',
    'RotationTool',
]

import vtk
import numpy as np
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa
# Import Helpers:
from ..base import FilterBase
from .. import _helpers



def PointsToPolyData(points):
    """Create ``vtkPolyData`` from a numpy array of XYZ points

    Return:
        vtkPolyData : points with point-vertex cells
    """
    __displayname__ = 'Points to PolyData'
    __type__ = 'filter'
    if points.ndim != 2:
        points = points.reshape((-1, 3))

    npoints = points.shape[0]

    # Make VTK cells array
    cells = np.hstack((np.ones((npoints, 1)),
                       np.arange(npoints).reshape(-1, 1)))
    cells = np.ascontiguousarray(cells, dtype=np.int64)
    vtkcells = vtk.vtkCellArray()
    vtkcells.SetCells(npoints, nps.numpy_to_vtkIdTypeArray(cells, deep=True))

    # Convert points to vtk object
    pts = vtk.vtkPoints()
    pts.SetData(nps.numpy_to_vtk(points))

    # Create polydata
    pdata = vtk.vtkPolyData()
    pdata.SetPoints(pts)
    pdata.SetVerts(vtkcells)
    return pdata


###############################################################################
#---- LatLon to Cartesian ----#
def latLonTableToCartesian(pdi, arrlat, arrlon, arralt, radius=6371.0, pdo=None):
    """**WORK IN PROGRESS**
    """
    # TODO: This is very poorly done
    # TODO: filter works but assumes a spherical earth wich is VERY wrong
    # NOTE: Mismatches the vtkEarth Source however so we gonna keep it this way
    raise _helpers.PVGeoError('latLonTableToCartesian() not currently implemented.')
    if pdo is None:
        pdo = vtk.vtkPolyData()
    #pdo.DeepCopy(pdi)
    wpdo = dsa.WrapDataObject(pdo)
    import sys
    sys.path.append('/Users/bane/miniconda3/lib/python3.6/site-packages/')
    import utm

    # Get the input arrays
    (namelat, fieldlat) = arrlat[0], arrlat[1]
    (namelon, fieldlon) = arrlon[0], arrlon[1]
    (namealt, fieldalt) = arralt[0], arralt[1]
    wpdi = dsa.WrapDataObject(pdi)
    lat = _helpers.getArray(wpdi, fieldlat, namelat)
    lon = _helpers.getArray(wpdi, fieldlon, namelon)
    alt = _helpers.getArray(wpdi, fieldalt, namealt)
    if len(lat) != len(lon) or len(lat) != len(alt):
        raise _helpers.PVGeoError('Latitude, Longitude, and Altitude arrays must be same length.')

    coords = np.empty((len(lat),3))

    for i in range(len(lat)):
        (e, n, zN, zL) = utm.from_latlon(lat[i], lon[i])
        coords[i,0] = e
        coords[i,1] = n
    coords[:,2] = alt

    pdo.ShallowCopy(PointsToPolyData(coords))
    # Add other arrays to output appropriately
    pdo = _helpers.copyArraysToPointData(pdi, pdo, fieldlat)

    return pdo

###############################################################################


class RotationTool(object):
    """A class that holds a set of methods/tools for performing and estimating coordinate rotations.
    """
    __displayname__ = 'Rotation Tool'
    __type__ = 'filter'
    def __init__(self, decimals=6):
        # Parameters
        self.RESOLUTION = np.pi / 3200.0
        self.DECIMALS = decimals

    @staticmethod
    def _GetRotationMatrix(theta):
        return np.array([
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)]])

    @staticmethod
    def RotateAround(pts, theta, origin):
        """Rotate points around an origins given an anlge on the XY plane
        """
        xarr, yarr = pts[:,0], pts[:,1]
        ox, oy = origin[0], origin[1]
        qx = ox + np.cos(theta) * (xarr - ox) - np.sin(theta) * (yarr - oy)
        qy = oy + np.sin(theta) * (xarr - ox) + np.cos(theta) * (yarr - oy)
        return np.vstack((qx, qy)).T

    @staticmethod
    def Rotate(pts, theta):
        """Rotate points around (0,0,0) given an anlge on the XY plane
        """
        rot = RotationTool._GetRotationMatrix(theta)
        return pts.dot(rot)

    @staticmethod
    def DistanceBetween(pts):
        """Gets the distance between two points
        """
        return np.sqrt((pts[0,0] - pts[1,0])**2 + (pts[0,1] - pts[1,1])**2)

    @staticmethod
    def CosBetween(pts):
        xdiff = abs(pts[0,0] - pts[1,0])
        dist = RotationTool.DistanceBetween(pts)
        return np.arccos(xdiff/dist)

    @staticmethod
    def SinBetween(pts):
        ydiff = abs(pts[0,1] - pts[1,1])
        dist = RotationTool.DistanceBetween(pts)
        return np.arcsin(ydiff/dist)


    # def _ConvergeAngle2(self, pt1, pt2):
    #     """internal use only: pts should only be a two neighboring points"""
    #     # Make the theta range up to 90 degrees to rotate points through
    #     #- angles = [0.0, 90.0)
    #     angles = np.arange(0.0, np.pi/2, self.RESOLUTION)
    #     pts = self.Rotate(np.vstack((pt1, pt2)), angles)
    #     # Get the angles between the points
    #     c = self.CosBetween(pts)
    #     dist = self.DistanceBetween(pts)
    #
    #     # Find angles that satisfy grid conditions
    #     xidx = np.argwhere(abs(c - np.pi/2.0) < (1 * 10**-self.DECIMALS))
    #     yidx = np.argwhere(abs(c - 0.0) < (1 * 10**-self.DECIMALS))
    #     if len(xidx) == 1 and len(yidx) == 0:
    #         return 0, np.pi/2-angles[xidx], dist[xidx]
    #         #return angles[xidx[0][0]]
    #     elif len(yidx) == 1 and len(xidx) == 0:
    #         return 1, np.pi/2-angles[yidx], dist[yidx]
    #     else:
    #         raise _helpers.PVGeoError('No angle found')


    def _ConvergeAngle(self, pt1, pt2):
        """Internal use only: pts should only be a two neighboring points.
        """
        # Make the theta range up to 90 degrees to rotate points through
        #- angles = [0.0, 90.0)
        angles = np.arange(0.0, np.pi/2, self.RESOLUTION)
        nang = len(angles) # Number of rotations

        pts = np.vstack((pt1, pt2))

        for i in range(nang):
            # TODO: eliminate this for loop
            temp = self.Rotate(pts, angles[i])
            # Check the angles between the points
            c = self.CosBetween(temp)
            dist = self.DistanceBetween(temp)
            if abs(c - np.pi/2.0) < (1 * 10**-self.DECIMALS):
                return 0, angles[i], dist
            if abs(c - 0.0) < (1 * 10**-self.DECIMALS):
                return 1, angles[i], dist
        # Uh-oh... we reached here...
        #- lets try again with lower precision
        self.DECIMALS -= 1
        if self.DECIMALS < 0:
            self.DECIMALS = 0
            raise _helpers.PVGeoError('No angle found. Precision too low/high.')
        return self._ConvergeAngle(pt1, pt2)


    def _EstimateAngleAndSpacing(self, pts, sample=0.10):
        """internal use only
        """
        from scipy.spatial import cKDTree # NOTE: Must have SciPy in ParaView
        # Creat the indexing range for searching the points:
        num = len(pts)
        rng = np.linspace(0, num-1, num=num, dtype=int)
        N = int(num*(1.0-sample))
        rng = rng[0::N]
        angles = np.zeros(len(rng))
        tree = cKDTree(pts)
        distances = [[],[]]

        # Get angles
        idx = 0
        for i in rng:
            # Find nearest point
            dist, ptsi = tree.query(pts[i], k=2)
            pt1 = pts[ptsi[0]]
            pt2 = pts[ptsi[1]]
            ax, angles[idx], dist = self._ConvergeAngle(pt1, pt2)
            distances[ax].append(dist)
            idx += 1

        angle = np.average(np.unique(angles))
        dx = np.unique(distances[0])
        dy = np.unique(distances[1])
        if len(dx) == 0:
            dx = dy
        elif len(dy) == 0:
            dy = dx
        # Now make sure only one angle was found
        # if len(angles) > 1:
        #     # This algorithm assumes the input data has uniform spacing throughout.
        #     # If multiple angles were found then that assumption must be False making the input data invalid.
        #     raise _helpers.PVGeoError('More than one angle recovered: Input data is invalid. Seek help with <info@pvgeo.org>.')
        return angle, dx[0], dy[0]


    def EstimateAndRotate(self, x, y, z):
        """A method to estimate the rotation of a set of points and correct that rotation on the XY plane
        """
        assert(len(x) == len(y) == len(z))
        idxs = np.argwhere(z == z[0])
        pts = np.hstack((x[idxs], y[idxs]))
        angle, dx, dy = self._EstimateAngleAndSpacing(pts)
        inv = self.Rotate(np.vstack((x, y)).T, angle)
        return inv[:,0], inv[:,1], z, dx, dy, angle



#---- Coordinate Rotations ----#

class RotatePoints(FilterBase):
    """Rotates XYZ coordinates in `vtkPolyData` around an origin at a given angle on the XY plane.
    """
    __displayname__ = 'Rotate Points'
    __type__ = 'filter'
    def __init__(self, angle=45.0, origin=[0.0, 0.0], useCorner=True):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkPolyData',
            nOutputPorts=1, outputType='vtkPolyData')
        # Parameters
        self.__angle = angle
        self.__origin = origin
        self.__useCorner = useCorner

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output.
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        # Get the Points over the NumPy interface
        wpdi = dsa.WrapDataObject(pdi) # NumPy wrapped input
        points = np.array(wpdi.Points) # New NumPy array of poins so we dont destroy input
        origin = self.__origin
        if self.__useCorner:
            idx = np.argmin(points[:,0])
            origin = [points[idx,0], points[idx,1]]
        points[:,0:2] = RotationTool.RotateAround(points[:,0:2], self.__angle, origin)
        pdo.DeepCopy(pdi)
        pts = pdo.GetPoints()
        for i in range(len(points)):
            pts.SetPoint(i, points[i])
        return 1

    def SetRotationDegrees(self, theta):
        """Sets the rotational angle in degrees.
        """
        theta = np.deg2rad(theta)
        if self.__angle != theta:
            self.__angle = theta
            self.Modified()

    def SetOrigin(self, xo, yo):
        """Sets the origin to perform the rotate around.
        """
        if self.__origin != [xo, yo]:
            self.__origin = [xo, yo]
            self.Modified()

    def SetUseCorner(self, flag):
        """A flag to use a corner of the input data set as the rotational origin.
        """
        if self.__useCorner != flag:
            self.__useCorner = flag
            self.Modified()



###############################################################################

class ExtractPoints(FilterBase):
    """Extracts XYZ coordinates and point/cell data from an input ``vtkDataSet``
    """
    __displayname__ = 'Extract Points'
    __type__ = 'filter'
    def __init__(self):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkDataSet',
            nOutputPorts=1, outputType='vtkPolyData')

    def RequestData(self, request, inInfo, outInfo):
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        #### Perfrom task ####
        # Get the Points over the NumPy interface
        wpdi = dsa.WrapDataObject(pdi) # NumPy wrapped input
        if not hasattr(wpdi, 'Points'):
            raise _helpers.PVGeoError('Input data object does not have XYZ points.')
        points = np.array(wpdi.Points) # New NumPy array of poins so we dont destroy input
        # Now transfer data
        f = vtk.vtkCellDataToPointData()
        f.SetInputData(pdi)
        f.Update()
        d = f.GetOutput()
        pdo.ShallowCopy(PointsToPolyData(points))
        _helpers.copyArraysToPointData(d, pdo, 0) # 0 is point data
        return 1



__all__ = [
    'EarthSource',
]

import vtk
# Import Helpers:
from ..base import AlgorithmBase
from .. import _helpers


class EarthSource(AlgorithmBase):
    """A simple data source to produce a ``vtkEarthSource``
    """
    __displayname__ = 'GSLib Table Reader'
    __type__ = 'source'
    def __init__(self, radius=6371.0):
        AlgorithmBase.__init__(self,
            nInputPorts=0,
            nOutputPorts=1, outputType='vtkPolyData')
        self.__radius = radius

    def RequestData(self, request, inInfo, outInfo):
        pdo = self.GetOutputData(outInfo, 0)
        earth = vtk.vtkEarthSource()
        earth.SetRadius(self.__radius)
        earth.OutlineOn()
        earth.Update()
        pdo.ShallowCopy(earth.GetOutput())
        return 1

    def SetRadius(self, radius):
        if self.__radius != radius:
            self.__radius = radius
            self.Modified()



__all__ = [
    'GSLibReader',
]

import numpy as np
from vtk.util import numpy_support as nps
# Import Helpers:
from .. import _helpers
from ..readers import DelimitedTextReader

class GSLibReader(DelimitedTextReader):
    """Reads a GSLIB file format to a ``vtkTable``. The GSLIB file format has headers lines followed by the data as a space delimited ASCI file (this filter is set up to allow you to choose any single character delimiter). The first header line is the title and will be printed to the console. This line may have the dimensions for a grid to be made of the data. The second line is the number (n) of columns of data. The next n lines are the variable names for the data in each column. You are allowed up to ten characters for the variable name. The data follow with a space between each field (column).
    """
    __displayname__ = 'GSLib Table Reader'
    __type__ = 'reader'
    def __init__(self, outputType='vtkTable', **kwargs):
        DelimitedTextReader.__init__(self, outputType=outputType, **kwargs)
        self.SetDelimiter(kwargs.get('delimiter', ' '))
        # These are attributes the derived from file contents:
        self.__header = None

    def _ExtractHeader(self, content):
        self.__header = content[0]
        try:
            num = int(content[1]) # number of data columns
        except ValueError:
            raise _helpers.PVGeoError('This file is not in proper GSLIB format.')
        titles = [ln.rstrip('\r\n') for ln in content[2:2+num]]
        return titles, content[2 + num::]


    #### Seters and Geters ####

    def GetFileHeader(self):
        """Returns the file header. If file hasn't been read, returns ``None``
        """
        return self.__header



__all__ = [
    'ReverseImageDataAxii',
]

from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa
import numpy as np
# Import Helpers:
from ..base import FilterBase
from .. import _helpers


#---- Reverse Grid Axii ----#


class ReverseImageDataAxii(FilterBase):
    """This filter will flip ``vtkImageData`` on any of the three cartesian axii. A checkbox is provided for each axis on which you may desire to flip the data."""
    __displayname__ = 'Reverse Image Data Axii'
    __type__ = 'filter'
    def __init__(self, axes=[True, True, True]):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkImageData',
            nOutputPorts=1, outputType='vtkImageData')
        self.__axes = axes[::-1] # Z Y X (FORTRAN)

    def _ReverseGridAxii(self, idi, ido):
        # Copy over input to output to be flipped around
        # Deep copy keeps us from messing with the input data
        ox, oy, oz = idi.GetOrigin()
        ido.SetOrigin(ox, oy, oz)
        sx, sy, sz = idi.GetSpacing()
        ido.SetSpacing(sx, sy, sz)
        ext = idi.GetExtent()
        nx, ny, nz = ext[1]+1, ext[3]+1, ext[5]+1
        ido.SetDimensions(nx, ny, nz)

        widi = dsa.WrapDataObject(idi)
        # Iterate over all array in the PointData
        for j in range(idi.GetPointData().GetNumberOfArrays()):
            # Go through each axis and rotate if needed
            arr = widi.PointData[j]
            arr = np.reshape(arr, (nz,ny,nx))
            for i in range(3):
                if self.__axes[i]:
                    arr = np.flip(arr, axis=i)
            # Now add that data array to the output
            data = nps.numpy_to_vtk(num_array=arr.flatten(), deep=True)
            data.SetName(idi.GetPointData().GetArrayName(j))
            ido.GetPointData().AddArray(data)

        # Iterate over all array in the CellData
        for j in range(idi.GetCellData().GetNumberOfArrays()):
            # Go through each axis and rotate if needed
            arr = widi.CellData[j]
            arr = np.reshape(arr, (nz-1,ny-1,nx-1))
            for i in range(3):
                if self.__axes[i]:
                    arr = np.flip(arr, axis=i)
            # Now add that data array to the output
            data = nps.numpy_to_vtk(num_array=arr.flatten(), deep=True)
            data.SetName(idi.GetCellData().GetArrayName(j))
            ido.GetCellData().AddArray(data)

        return ido

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output.
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._ReverseGridAxii(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def SetFlipX(self, flag):
        """Set the filter to flip th input data along the X-axis
        """
        if self.__axes[2] != flag:
            self.__axes[2] = flag
            self.Modified()

    def SetFlipY(self, flag):
        """Set the filter to flip th input data along the Y-axis
        """
        if self.__axes[1] != flag:
            self.__axes[1] = flag
            self.Modified()

    def SetFlipZ(self, flag):
        """Set the filter to flip th input data along the Z-axis
        """
        if self.__axes[0] != flag:
            self.__axes[0] = flag
            self.Modified()



__all__ = [
    'TableToGrid',
]

import vtk
from vtk.util import numpy_support as nps
import numpy as np
# Import Helpers:
from ..base import FilterBase
from .. import _helpers


#---- Table To Grid Stuff ----#



class TableToGrid(FilterBase):
    """This filter takes a ``vtkTable`` object with columns that represent data to be translated (reshaped) into a 3D grid (2D also works, just set the third dimensions extent to 1). The grid will be a ``n1`` by ``n2`` by ``n3`` ``vtkImageData`` structure and an origin (south-west bottom corner) can be set at any xyz point. Each column of the ``vtkTable`` will represent a data attribute of the ``vtkImageData`` formed (essentially a uniform mesh). The SEPlib option allows you to unfold data that was packed in the SEPlib format where the most important dimension is z and thus the z data is d1 (``d1=z``, ``d2=y``, ``d3=x``). When using SEPlib, specify ``n1`` as the number of elements in the Z-direction, ``n2`` as the number of elements in the X-direction, and ``n3`` as the number of elements in the Y-direction (and so on for other parameters).

    Warning:
        **Work in progress**
    """
    __displayname__ = 'Table To Grid'
    __type__ = 'filter'
    def __init__(self, extent=[10, 10, 10], order='C', spacing=[1.0, 1.0, 1.0],
                 origin=[0.0, 0.0, 0.0], seplib=False, swapXY=False):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkTable',
            nOutputPorts=1, outputType='vtkImageData')
        self.__extent = extent # MUST BE SET
        self.__spacing = spacing
        self.__origin = origin
        self.__SEPlib = seplib
        self.__order = order
        self.__swapXY = swapXY


    @staticmethod
    def _unpack(arr, extent, order='C'):
        """This is a helper method that handles the initial unpacking of a data array.
        ParaView and VTK use Fortran packing so this is convert data saved in
        C packing to Fortran packing.
        """
        n1,n2,n3 = extent[0],extent[1],extent[2]
        if order == 'C':
            arr = np.reshape(arr, (n1,n2,n3))
            arr = np.swapaxes(arr,0,2)
            extent = np.shape(arr)
        elif order == 'F':
            # effectively doing nothing
            #arr = np.reshape(arr, (n3,n2,n1))
            return arr.flatten(), extent
        return arr.flatten(), extent

    @staticmethod
    def _rearangeSEPlib(arr, extent):
        """This is a helper method to swap axes when using SEPlib axial conventions.
        """
        n1,n2,n3 = extent[0],extent[1],extent[2]
        arr = np.reshape(arr, (n3,n2,n1))
        arr = np.swapaxes(arr,0,2)
        return arr.flatten(), (n1,n2,n3)

    @staticmethod
    def _transposeXY(arr, extent, SEPlib=False):
        """Transposes X and Y axes. Needed for PoroTomo project.
        """
        n1,n2,n3 = extent[0],extent[1],extent[2]
        arr = np.reshape(arr, (n1,n2,n3))
        if SEPlib:
            arr = np.swapaxes(arr,1,2)
            ext = (n1,n3,n2)
        else:
            # print('bef: ', np.shape(arr))
            arr = np.swapaxes(arr,0,1)
            # print('aft: ', np.shape(arr))
            ext = np.shape(arr)
        return arr.flatten(), ext

    @staticmethod
    def _refold(arr, extent, SEPlib=True, order='F', swapXY=False):
        """This is a helper method to handle grabbing a data array and make sure it is ready for VTK/Fortran ordering in ``vtkImageData``.
        """
        # Fold into 3D using extents. Packing dimensions should be in order extent
        arr, extent = TableToGrid._unpack(arr, extent, order=order)
        if SEPlib:
            arr, extent = TableToGrid._rearangeSEPlib(arr, extent)
        if swapXY:
            arr, extent = TableToGrid._transposeXY(arr, extent, SEPlib=SEPlib)
        return arr#, extent

    @staticmethod
    def RefoldIdx(SEPlib=True, swapXY=False):
        """Theses are indexing corrections to set the spacings and origin witht the correct axes after refolding.
        """
        if SEPlib:
            idx = (2,1,0)
            if swapXY:
                idx = (1,2,0)
        else:
            idx = (0,1,2)
            if swapXY:
                idx = (1,0,2)
        return idx

    def _TableToGrid(self, pdi, ido):
        """Converts a table of data arrays to vtkImageData given an extent to reshape that table.
        Each column in the table will be treated as seperate data arrays for the described data space.
        """
        cols = pdi.GetNumberOfColumns()
        rows = pdi.GetColumn(0).GetNumberOfTuples()

        idx = TableToGrid.RefoldIdx(SEPlib=self.__SEPlib, swapXY=self.__swapXY)
        nx,ny,nz = self.__extent[idx[0]],self.__extent[idx[1]],self.__extent[idx[2]]
        sx,sy,sz = self.__spacing[idx[0]],self.__spacing[idx[1]],self.__spacing[idx[2]]
        ox,oy,oz = self.__origin[idx[0]],self.__origin[idx[1]],self.__origin[idx[2]]
        # make sure dimensions work
        if (nx*ny*nz != rows):
            raise _helpers.PVGeoError('Total number of elements must remain %d. Check reshape dimensions (n1 by n2 by n3).' % (rows))

        ido.SetDimensions(nx, ny, nz)
        #ido.SetExtent(0,nx-1, 0,ny-1, 0,nz-1)
        ido.SetOrigin(ox, oy, oz)
        ido.SetSpacing(sx, sy, sz)

        # Add all columns of the table as arrays to the PointData
        for i in range(cols):
            c = pdi.GetColumn(i)
            name = c.GetName()
            arr = nps.vtk_to_numpy(c)
            arr = TableToGrid._refold(arr, self.__extent, SEPlib=self.__SEPlib, order=self.__order, swapXY=self.__swapXY)
            c = nps.numpy_to_vtk(num_array=arr, deep=True)
            c.SetName(name)
            #ido.GetCellData().AddArray(c) # Should we add here? flipper won't flip these...
            # Also, image data is built by points from numrows in table
            ido.GetPointData().AddArray(c)
            #scal = ido.GetPointData().GetArray(i)

        return ido

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        ido = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._TableToGrid(pdi, ido)
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        """Used by pipeline to set whole output extent.
        """
        # Setup the ImageData
        idx = TableToGrid.RefoldIdx(SEPlib=self.__SEPlib, swapXY=self.__swapXY)
        ext = [0, self.__extent[idx[0]]-1, 0,self.__extent[idx[1]]-1, 0,self.__extent[idx[2]]-1]
        info = outInfo.GetInformationObject(0)
        # Set WHOLE_EXTENT: This is absolutely necessary
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        return 1


    #### Setters / Getters ####


    def SetExtent(self, nx, ny, nz):
        """Set the extent of the output grid"""
        if self.__extent != [nx, ny, nz]:
            self.__extent = [nx, ny, nz]
            self.Modified()

    def SetSpacing(self, dx, dy, dz):
        """Set the spacing for the points along each axial direction
        """
        if self.__spacing != [dx, dy, dz]:
            self.__spacing = [dx, dy, dz]
            self.Modified()

    def SetOrigin(self, x0, y0, z0):
        """Set the origin of the output `vtkImageData`
        """
        if self.__origin != [x0, y0, z0]:
            self.__origin = [x0, y0, z0]
            self.Modified()

    def SetSEPlib(self, flag):
        """Set a flag to swap the axial order for the refold of the table
        """
        if self.__SEPlib != flag:
            self.__SEPlib = flag
            self.Modified()

    def SetOrder(self, order):
        """Set the reshape order (`'C'` or `'F'`)"""
        if self.__order != order:
            self.__order = order
            self.Modified()

    def SetSwapXY(self, flag):
        """Set a flag to swap the X and Y axii
        """
        if self.__swapXY != flag:
            self.__swapXY = flag
            self.Modified()



__all__ = [
    'SGeMSGridReader',
]

import numpy as np
from vtk.util import numpy_support as nps
import vtk

from .gslib import GSLibReader
from .. import _helpers


class SGeMSGridReader(GSLibReader):
    """Generates ``vtkImageData`` from the uniform grid defined in the inout file in the SGeMS grid format. This format is simply the GSLIB format where the header line defines the dimensions of the uniform grid.
    """
    __displayname__ = 'SGeMS Grid Reader'
    __type__ = 'reader'
    def __init__(self, origin=(0.0, 0.0, 0.0), spacing=(1.0, 1.0, 1.0), **kwargs):
        GSLibReader.__init__(self, outputType='vtkImageData', **kwargs)
        self.__extent = None
        self.__origin = origin
        self.__spacing = spacing

    def _ReadExtent(self):
        """Reads the input file for the SGeMS format to get output extents. Computationally inexpensive method to discover whole output extent.

        Return:
            tuple : This returns a tuple of the whole extent for the uniform grid to be made of the input file (0,n1-1, 0,n2-1, 0,n3-1). This output should be directly passed to set the whole output extent.

        """
        # Read first file... extent cannot vary with time
        # TODO: make more efficient to only reader header of file
        fileLines = self._GetFileContents(idx=0)
        h = fileLines[0+self.GetSkipRows()].split(self._GetDeli())
        try:
            n1,n2,n3 = int(h[0]), int(h[1]), int(h[2])
        except ValueError:
            raise _helpers.PVGeoError('File not in proper SGeMS Grid fromat.')
        return (0,n1-1, 0,n2-1, 0,n3-1)

    def _ExtractHeader(self, content):
        titles, content = GSLibReader._ExtractHeader(self, content)
        h = self.GetFileHeader().split(self._GetDeli())
        try:
            if self.__extent is None:
                self.__extent = (int(h[0]), int(h[1]), int(h[2]))
            elif self.__extent != (int(h[0]), int(h[1]), int(h[2])):
                raise _helpers.PVGeoError('Grid dimensions change in file time series.')
        except ValueError:
            raise _helpers.PVGeoError('File not in proper SGeMS Grid fromat.')
        return titles, content

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to get output data object for given time step. Constructs the ``vtkImageData``
        """
        # Get output:
        output = vtk.vtkImageData.GetData(outInfo)
        # Get requested time index
        i = _helpers.GetRequestedTime(self, outInfo)
        if self.NeedToRead():
            self._ReadUpFront()
        # Generate the data object
        n1, n2, n3 = self.__extent
        dx, dy, dz = self.__spacing
        ox, oy, oz = self.__origin
        output.SetDimensions(n1, n2, n3)
        output.SetExtent(0,n1-1, 0,n2-1, 0,n3-1)
        output.SetSpacing(dx, dy, dz)
        output.SetOrigin(ox, oy, oz)
        # Use table generater and convert because its easy:
        table = vtk.vtkTable()
        _helpers.placeArrInTable(self._GetRawData(idx=i), self.GetTitles(), table)
        # now get arrays from table and add to point data of pdo
        for i in range(table.GetNumberOfColumns()):
            output.GetPointData().AddArray(table.GetColumn(i))
            #TODO: maybe we ought to add the data as cell data
        del(table)
        return 1


    def RequestInformation(self, request, inInfo, outInfo):
        """Used by pipeline to set grid extents.
        """
        # Call parent to handle time stuff
        GSLibReader.RequestInformation(self, request, inInfo, outInfo)
        # Now set whole output extent
        ext = self._ReadExtent()
        info = outInfo.GetInformationObject(0)
        # Set WHOLE_EXTENT: This is absolutely necessary
        info.Set(vtk.vtkStreamingDemandDrivenPipeline.WHOLE_EXTENT(), ext, 6)
        return 1


    def SetSpacing(self, dx, dy, dz):
        """Set the spacing for each axial direction"""
        spac = (dx, dy, dz)
        if self.__spacing != spac:
            self.__spacing = spac
            self.Modified(readAgain=False)

    def SetOrigin(self, ox, oy, oz):
        """Set the origin corner of the grid"""
        origin = (ox, oy, oz)
        if self.__origin != origin:
            self.__origin = origin
            self.Modified(readAgain=False)



__all__ = [
    'CombineTables',
    'ReshapeTable',
]

import numpy as np
from vtk.util import numpy_support as nps
from vtk.numpy_interface import dataset_adapter as dsa
# Import Helpers:
from ..base import FilterBase
from .. import _helpers


###############################################################################

###############################################################################


class CombineTables(FilterBase):
    """Takes two tables and combines them if they have the same number of rows.
    """
    __displayname__ = 'Combine Tables'
    __type__ = 'filter'
    def __init__(self):
        FilterBase.__init__(self,
            nInputPorts=2, inputType='vtkTable',
            nOutputPorts=1, outputType='vtkTable')
        # Parameters... none

    # CRITICAL for multiple input ports
    def FillInputPortInformation(self, port, info):
        """Used by pipeline. Necessary when dealing with multiple input ports
        """
        # all are tables so no need to check port
        info.Set(self.INPUT_REQUIRED_DATA_TYPE(), "vtkTable")
        return 1


    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output
        """
        # Inputs from different ports:
        pdi0 = self.GetInputData(inInfo, 0, 0)
        pdi1 = self.GetInputData(inInfo, 1, 0)
        pdo = self.GetOutputData(outInfo, 0)

        pdo.DeepCopy(pdi0)

        # Get number of rows
        nrows = pdi0.GetNumberOfRows()
        nrows1 = pdi1.GetNumberOfRows()
        assert(nrows == nrows1)

        for i in range(pdi1.GetRowData().GetNumberOfArrays()):
            arr = pdi1.GetRowData().GetArray(i)
            pdo.GetRowData().AddArray(arr)
        return 1

    def Apply(self, table0, table1):
        self.SetInputDataObject(0, table0)
        self.SetInputDataObject(1, table1)
        self.Update()
        return self.GetOutput()


###############################################################################
#---- Reshape Table ----#

class ReshapeTable(FilterBase):
    """This filter will take a ``vtkTable`` object and reshape it. This filter essentially treats ``vtkTable``s as 2D matrices and reshapes them using ``numpy.reshape`` in a C contiguous manner. Unfortunately, data fields will be renamed arbitrarily because VTK data arrays require a name.
    """
    __displayname__ = 'Reshape Table'
    __type__ = 'filter'
    def __init__(self, **kwargs):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkTable',
            nOutputPorts=1, outputType='vtkTable')
        # Parameters
        self.__nrows = kwargs.get('nrows', 1)
        self.__ncols = kwargs.get('ncols', 1)
        self.__names = kwargs.get('names', [])
        self.__order = kwargs.get('order', 'F')

    def _Reshape(self, pdi, pdo):
        """Internal helper to perfrom the reshape
        """
        # Get number of columns
        cols = pdi.GetNumberOfColumns()
        # Get number of rows
        rows = pdi.GetColumn(0).GetNumberOfTuples()

        if len(self.__names) is not 0:
            num = len(self.__names)
            if num < self.__ncols:
                for i in range(num, self.__ncols):
                    self.__names.append('Field %d' % i)
            elif num > self.__ncols:
                raise _helpers.PVGeoError('Too many array names. `ncols` specified as %d and %d names given.' % (self.__ncols, num))
        else:
            self.__names = ['Field %d' % i for i in range(self.__ncols)]

        # Make a 2D numpy array and fill with data from input table
        data = np.empty((rows,cols))
        for i in range(cols):
            c = pdi.GetColumn(i)
            data[:,i] = nps.vtk_to_numpy(c)

        if ((self.__ncols*self.__nrows) != (cols*rows)):
            raise _helpers.PVGeoError('Total number of elements must remain %d. Check reshape dimensions.' % (cols*rows))

        # Use numpy.reshape() to reshape data NOTE: only 2D because its a table
        # NOTE: column access of this reshape is not contigous
        data = np.array(np.reshape(data.flatten(), (self.__nrows,self.__ncols), order=self.__order))
        pdo.SetNumberOfRows(self.__nrows)

        # Add new array to output table and assign incremental names (e.g. Field0)
        for i in range(self.__ncols):
            # Make a contigous array from the column we want
            col = np.array(data[:,i])
            # allow type to be determined by input
            insert = nps.numpy_to_vtk(num_array=col, deep=True) # array_type=vtk.VTK_FLOAT
            # VTK arrays need a name. Set arbitrarily
            insert.SetName(self.__names[i])
            #pdo.AddColumn(insert) # these are not getting added to the output table
            # ... work around:
            pdo.GetRowData().AddArray(insert) # NOTE: this is in the FieldData

        return pdo

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        self._Reshape(pdi, pdo)
        return 1


    #### Seters and Geters ####

    def SetNames(self, names):
        """Set names using a semicolon (;) seperated string or a list of strings

        Args:
            names (string): a string of data array names for the reshaped table using a semicolon (;) to spearate
        """
        # parse the names (a semicolon seperated list of names)
        if isinstance(names, str):
            names = names.split(';')
        if self.__names != names:
            self.__names = names
            self.Modified()

    def AddName(self, name):
        """Use to append a name to the list of data array names for the output table
        """
        self.__names.append(name)
        self.Modified()

    def GetNames(self):
        return self.__names

    def SetNumberOfColumns(self, ncols):
        """Set the number of columns for the output ``vtkTable``
        """
        if isinstance(ncols, float):
            ncols = int(ncols)
        if self.__ncols != ncols:
            self.__ncols = ncols
            self.Modified()

    def SetNumberOfRows(self, nrows):
        """Set the number of rows for the output ``vtkTable``
        """
        if isinstance(nrows, float):
            nrows = int(nrows)
        if self.__nrows != nrows:
            self.__nrows = nrows
            self.Modified()

    def SetOrder(self, order):
        """Set the reshape order (``'C'`` of ``'F'``)
        """
        if self.__order != order:
            self.__order = order
            self.Modified()



__all__ = [
    'AlgorithmBase',
    'ReaderBase',
    'FilterPreserveTypeBase',
    'TwoFileReaderBase',
]


__displayname__ = 'Base Classes'

from . import _helpers

# Outside Imports:
import vtk.util.vtkAlgorithm as valg #import VTKPythonAlgorithmBase
import numpy as np

###############################################################################

class AlgorithmBase(valg.VTKPythonAlgorithmBase):
    """This is a base class to add convienace methods to the
    ``VTKPythonAlgorithmBase`` for all algorithms implemented in ``PVGeo``.
    We implement our algorithms in this manner to harness all of the backend support that the ``VTKPythonAlgorithmBase`` class provides for integrating custom algorithms on a VTK pipeline. All of the pipeline methods for setting inputs, getting outputs, making requests are handled by the super classes. For more information on what functionality is available, check out the VTK Docs for the `vtkAlgorithm`_ and then check out the following blog posts:

    * `vtkPythonAlgorithm is great`_
    * A VTK pipeline primer `(part 1)`_, `(part 2)`_, and `(part 3)`_

    .. _vtkAlgorithm: https://www.vtk.org/doc/nightly/html/classvtkAlgorithm.html
    .. _vtkPythonAlgorithm is great: https://blog.kitware.com/vtkpythonalgorithm-is-great/
    .. _(part 1): https://blog.kitware.com/a-vtk-pipeline-primer-part-1/
    .. _(part 2): https://blog.kitware.com/a-vtk-pipeline-primer-part-2/
    .. _(part 3): https://blog.kitware.com/a-vtk-pipeline-primer-part-3/
    """
    __displayname__ = 'Algorithm Base'
    __type__ = 'base'

    def __init__(self,
                nInputPorts=1, inputType='vtkDataSet',
                nOutputPorts=1, outputType='vtkTable'):
        valg.VTKPythonAlgorithmBase.__init__(self,
            nInputPorts=nInputPorts, inputType=inputType,
            nOutputPorts=nOutputPorts, outputType=outputType)
        # Add error handler to make errors easier to deal with
        self.__errorObserver = _helpers.ErrorObserver()
        self.__errorObserver.MakeObserver(self)

    def GetOutput(self, port=0):
        """A conveience method to get the output data object of this ``PVGeo`` algorithm.
        """
        return self.GetOutputDataObject(port)

    def ErrorOccurred(self):
        """A conveience method for handling errors on the VTK pipeline

        Return:
            bool: true if an error has ovvured since last checked
        """
        return self.__errorObserver.ErrorOccurred()

    def ErrorMessage(self):
        """A conveience method to print the error message.
        """
        return self.__errorObserver.ErrorMessage()


###############################################################################

# Base Reader
class ReaderBase(AlgorithmBase):
    """A base class for inherrited functionality common to all reader algorithms
    """
    __displayname__ = 'Reader Base'
    __type__ = 'base'
    def __init__(self, nOutputPorts=1, outputType='vtkTable', **kwargs):
        AlgorithmBase.__init__(self,
            nInputPorts=0,
            nOutputPorts=nOutputPorts, outputType=outputType,
            **kwargs)
        # Attributes are namemangled to ensure proper setters/getters are used
        # For the VTK/ParaView pipeline
        self.__dt = kwargs.get('dt', 1.0)
        self.__timesteps = None
        # For the reader
        self.__fileNames = kwargs.get('filenames', [])
        # To know whether or not the read needs to perform
        self.__needToRead = True

    def NeedToRead(self, flag=None):
        """Ask self if the reader needs to read the files again.

        Args:
            flag (bool): Set the read status

        Return:
            bool: the status of the reader.
        """
        if flag is not None and isinstance(flag, (bool, int)):
            self.__needToRead = flag
        return self.__needToRead

    def Modified(self, readAgain=True):
        """Call modified if the files needs to be read again again
        """
        if readAgain: self.__needToRead = readAgain
        AlgorithmBase.Modified(self)

    def _UpdateTimeSteps(self):
        """For internal use only: appropriately sets the timesteps.
        """
        self.__timesteps = _helpers.UpdateTimeSteps(self, self.__fileNames, self.__dt)
        return 1

    #### Algorithm Methods ####

    def RequestInformation(self, request, inInfo, outInfo):
        """This is a conveience method that should be overwritten when needed.
        This will handle setting the timesteps appropriately based on the number
        of file names when the pipeline needs to know the time information.
        """
        self._UpdateTimeSteps()
        return 1

    #### Methods for performing the read ####
    # These are meant to be overwritten by child classes

    def _GetFileContents(self, idx=None):
        raise NotImplementedError()

    def _ReadUpFront(self):
        raise NotImplementedError()

    def _GetRawData(self, idx=0):
        raise NotImplementedError()

    #### Seters and Geters ####

    def GetTimestepValues(self):
        """Use this in ParaView decorator to register timesteps on the pipeline.
        """
        return self.__timesteps.tolist() if self.__timesteps is not None else None

    def SetTimeDelta(self, dt):
        """An advanced property to set the time step in seconds.
        """
        if dt != self.__dt:
            self.__dt = dt
            self.Modified()

    def ClearFileNames(self):
        """Use to clear file names of the reader.

        Note:
            * This does not set the reader to need to read again as there are no files to read.
        """
        self.__fileNames = []

    def AddFileName(self, fname):
        """Use to set the file names for the reader. Handles singlt string or list of strings.
        Args:
            fname (str): The absolute file name with path to read.
        """
        if fname is None:
            return # do nothing if None is passed by a constructor on accident
        if isinstance(fname, list):
            for f in fname:
                self.AddFileName(f)
        elif fname not in self.__fileNames:
            self.__fileNames.append(fname)
        self.Modified()

    def GetFileNames(self, idx=None):
        """Returns the list of file names or given and index returns a specified
        timestep's filename.
        """
        if idx is None:
            return self.__fileNames
        return self.__fileNames[idx]

    def Apply(self, fname):
        """Given a file name (or list of file names), perfrom the read"""
        self.AddFileName(fname)
        self.Update()
        return self.GetOutput()

###############################################################################

# Base filter to preserve input data type
class FilterBase(AlgorithmBase):
    """A base class for implementing filters which holds several convienace methods"""
    __displayname__ = 'Filter Base'
    __type__ = 'base'
    def __init__(self,
        nInputPorts=1, inputType='vtkDataSet',
        nOutputPorts=1, outputType='vtkPolyData'):
        AlgorithmBase.__init__(self,
            nInputPorts=nInputPorts, inputType=inputType,
            nOutputPorts=nOutputPorts, outputType=outputType)

    def Apply(self, inputDataObject):
        self.SetInputDataObject(inputDataObject)
        self.Update()
        return self.GetOutput()



###############################################################################

# Base filter to preserve input data type
class FilterPreserveTypeBase(FilterBase):
    """A Base class for implementing filters that preserve the data type of
    their arbitrary input.
    """
    __displayname__ = 'Filter Preserve Type Base'
    __type__ = 'base'
    def __init__(self):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkDataObject',
            nOutputPorts=1)

    # THIS IS CRUCIAL to preserve data type through filter
    def RequestDataObject(self, request, inInfo, outInfo):
        """There is no need to overwrite this. This method lets the pipeline
        know that the algorithm will dynamically decide the output data type
        based in the input data type.
        """
        self.OutputType = self.GetInputData(inInfo, 0, 0).GetClassName()
        self.FillOutputPortInformation(0, outInfo.GetInformationObject(0))
        return 1

###############################################################################

# Two File Reader Base
class TwoFileReaderBase(AlgorithmBase):
    """A base clase for readers that need to handle two input files.
    One meta-data file and a series of data files.
    """
    __displayname__ = 'Two File Reader Base'
    __type__ = 'base'
    def __init__(self, nOutputPorts=1, outputType='vtkUnstructuredGrid', **kwargs):
        AlgorithmBase.__init__(self,
            nInputPorts=0,
            nOutputPorts=nOutputPorts, outputType=outputType)
        self.__dt = kwargs.get('dt', 1.0)
        self.__timesteps = None
        self.__meshFileName = kwargs.get('meshfile', None) # Can only be one!
        modfiles = kwargs.get('modelfiles', []) # Can be many (single attribute, manytimesteps)
        if isinstance(modfiles, str):
            modfiles = [modfiles]
        self.__modelFileNames = modfiles
        self.__needToReadMesh = True
        self.__needToReadModels = True


    def __UpdateTimeSteps(self):
        """For internal use only
        """
        if len(self.__modelFileNames) > 0:
            self.__timesteps = _helpers.UpdateTimeSteps(self, self.__modelFileNames, self.__dt)
        return 1

    def NeedToReadMesh(self, flag=None):
        """Ask self if the reader needs to read the mesh file again.

        Args:
            flag (bool): set the status of the reader for mesh files.
        """
        if flag is not None and isinstance(flag, (bool, int)):
            self.__needToReadMesh = flag
        return self.__needToReadMesh

    def NeedToReadModels(self, flag=None):
        """Ask self if the reader needs to read the model files again.

        Args:
            flag (bool): set the status of the reader for model files.
        """
        if flag is not None and isinstance(flag, (bool, int)):
            self.__needToReadModels = flag
        return self.__needToReadModels

    def Modified(self, readAgainMesh=True, readAgainModels=True):
        """Call modified if the files needs to be read again again

        Args:
            readAgainMesh (bool): set the status of the reader for mesh files.
            readAgainModels (bool): set the status of the reader for model files.
        """
        if readAgainMesh: self.NeedToReadMesh(flag=readAgainMesh)
        if readAgainModels: self.NeedToReadModels(flag=readAgainModels)
        return AlgorithmBase.Modified(self)

    def RequestInformation(self, request, inInfo, outInfo):
        self.__UpdateTimeSteps()
        return 1


    #### Seters and Geters ####


    @staticmethod
    def HasModels(modelfiles):
        """A convienance method to see if a list contatins models filenames.
        """
        if isinstance(modelfiles, list):
            return len(modelfiles) > 0
        return modelfiles is not None

    def ThisHasModels(self):
        """Ask self if the reader has model filenames set.
        """
        return TwoFileReaderBase.HasModels(self.__modelFileNames)

    def GetTimestepValues(self):
        """Use this in ParaView decorator to register timesteps
        """
        return self.__timesteps.tolist() if self.__timesteps is not None else None

    def SetTimeDelta(self, dt):
        """An advanced property for the time step in seconds.
        """
        if dt != self.__dt:
            self.__dt = dt
            self.Modified(readAgainMesh=False, readAgainModels=False)

    def ClearMesh(self):
        """Use to clear mesh file name
        """
        self.__meshFileName = None
        self.Modified(readAgainMesh=True, readAgainModels=False)

    def ClearModels(self):
        """Use to clear data file names
        """
        self.__modelFileNames = []
        self.Modified(readAgainMesh=False, readAgainModels=True)

    def SetMeshFileName(self, fname):
        """Set the mesh file name.
        """
        if self.__meshFileName != fname:
            self.__meshFileName = fname
            self.Modified(readAgainMesh=True, readAgainModels=False)

    def AddModelFileName(self, fname):
        """Use to set the file names for the reader. Handles single string or list of strings.

        Args:
            fname (str or list(str)): the file name(s) to use for the model data.
        """
        if fname is None:
            return # do nothing if None is passed by a constructor on accident
        if isinstance(fname, list):
            for f in fname:
                self.AddModelFileName(f)
            self.Modified(readAgainMesh=False, readAgainModels=True)
        elif fname not in self.__modelFileNames:
            self.__modelFileNames.append(fname)
            self.Modified(readAgainMesh=False, readAgainModels=True)
        return 1

    def GetModelFileNames(self, idx=None):
        """Returns the list of file names or given and index returns a specified
        timestep's filename.
        """
        if idx is None or not self.ThisHasModels():
            return self.__modelFileNames
        return self.__modelFileNames[idx]

    def GetMeshFileName(self):
        return self.__meshFileName

    def Apply(self):
        """Perfrom the read with parameters/file names set during init or by setters"""
        self.Update()
        return self.GetOutput()



"""PVGeo: an open-source python package for geoscientific visualization in VTK
and ParaView.
"""

import setuptools

__version__ = '1.1.10'

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="PVGeo",
    version=__version__,
    author="Bane Sullivan",
    author_email="info@pvgeo.org",
    description="Geoscientific visualization tools for VTK and ParaView",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OpenGeoVis/PVGeo",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.10',
        'scipy>=1.1',
        #'vtk>=8.1',
        'colour-runner==0.0.5',
        'codecov==2.0.15',
    ],
    classifiers=(
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Scientific/Engineering :: GIS',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
    ),
)



# Import base classes
from .base import *

# Import Suites
from . import filters
from . import grids
from . import gslib
from . import model_build
from . import readers
#TODO: from . import tunneling
from . import ubc

from . import _helpers
from .tester import test

__author__ = 'Bane Sullivan'
__license__ = 'BSD-3-Clause'
__copyright__ = '2018, Bane Sullivan'
__version__ = '1.1.10'
__displayname__ = 'PVGeo'



paraview_plugin_version = '1.1.10'
# This is module to import. It provides VTKPythonAlgorithmBase, the base class
# for all python-based vtkAlgorithm subclasses in VTK and decorators used to
# 'register' the algorithm with ParaView along with information about UI.
from paraview.util.vtkAlgorithm import *

# Helpers:
from PVGeo import _helpers
# Classes to Decorate
from PVGeo.ubc import *

#### GLOBAL VARIABLES ####
MENU_CAT = 'PVGeo: UBC Mesh Tools'

MESH_EXTS = 'mesh msh dat txt text'
TMESH_DESC = 'PVGeo: UBC Mesh 2D/3D Two-File Format'


@smproxy.reader(name="PVGeoTensorMeshReader",
       label="PVGeo: UBC Tensor Mesh Reader",
       extensions=MESH_EXTS,
       file_description=TMESH_DESC)
@smhint.xml('''<RepresentationType view="RenderView" type="Surface With Edges" />''')
class PVGeoTensorMeshReader(TensorMeshReader):
    def __init__(self):
        TensorMeshReader.__init__(self)


    #### Seters and Geters ####

    @smproperty.xml('''
        <StringVectorProperty
            panel_visibility="advanced"
            name="MeshFile"
            label="File Name Mesh"
            command="SetMeshFileName"
            animateable="1"
            clean_command="ClearMesh"
            number_of_elements="1">
            <FileListDomain name="meshfile"/>
            <Documentation>This is the mesh file for a 2D or 3D UBC Mesh grid. This plugin only allows ONE mesh to be defined.</Documentation>
        </StringVectorProperty>''')
    def SetMeshFileName(self, fname):
        TensorMeshReader.SetMeshFileName(self, fname)

    @smproperty.xml('''
        <StringVectorProperty
          panel_visibility="default"
          name="ModelFiles"
          label="File Name(s) Model"
          command="AddModelFileName"
          animateable="1"
          repeat_command="1"
          clean_command="ClearModels"
          number_of_elements="1">
          <FileListDomain name="modelfiles"/>
          <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
        </StringVectorProperty>''')
    def AddModelFileName(self, fname):
        """Use to set the file names for the reader. Handles singlt string or list of strings."""
        TensorMeshReader.AddModelFileName(self, fname)


    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def SetTimeDelta(self, dt):
        TensorMeshReader.SetTimeDelta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        """This is critical for registering the timesteps"""
        return TensorMeshReader.GetTimestepValues(self)

    @smproperty.stringvector(name='DataName', default_values='Data')
    def SetDataName(self, name):
        TensorMeshReader.SetDataName(self, name)

@smproxy.filter(name="PVGeoTensorMeshAppender",
       label="Append Model To UBC Tensor Mesh")
@smhint.xml('''<RepresentationType view="RenderView" type="Surface With Edges" />''')
@smhint.xml('<ShowInMenu category="%s"/>' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkRectilinearGrid"], composite_data_supported=False)
class PVGeoTensorMeshAppender(TensorMeshAppender):
    """This assumes the input vtkRectilinearGrid has already handled the timesteps"""
    def __init__(self):
        TensorMeshAppender.__init__(self)

    @smproperty.xml('''
        <StringVectorProperty
          panel_visibility="default"
          name="ModelFiles"
          label="File Name(s) Model"
          command="AddModelFileName"
          animateable="1"
          repeat_command="1"
          clean_command="ClearModels"
          number_of_elements="1">
          <FileListDomain name="modelfiles"/>
          <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
        </StringVectorProperty>''')
    def AddModelFileName(self, fname):
        """Use to set the file names for the reader. Handles single string or list of strings."""
        TensorMeshAppender.AddModelFileName(self, fname)

    @smproperty.stringvector(name='DataName', default_values='Appended Data')
    def SetDataName(self, name):
        TensorMeshAppender.SetDataName(self, name)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        """This is critical for registering the timesteps"""
        return TensorMeshAppender.GetTimestepValues(self)



#------------------------------------------------------------------------------
# Read OcTree Mesh
#------------------------------------------------------------------------------


@smproxy.reader(name="PVGeoUBCOcTreeMeshReader",
       label="PVGeo: UBC OcTree Mesh Reader",
       extensions=MESH_EXTS,
       file_description='PVGeo: UBC OcTree Mesh')
@smhint.xml('''<RepresentationType view="RenderView" type="Surface With Edges" />''')
class PVGeoUBCOcTreeMeshReader(OcTreeReader):
    def __init__(self):
        OcTreeReader.__init__(self)


    #### Seters and Geters ####

    @smproperty.xml('''
        <StringVectorProperty
            panel_visibility="advanced"
            name="MeshFile"
            label="File Name Mesh"
            command="SetMeshFileName"
            animateable="1"
            clean_command="ClearMesh"
            number_of_elements="1">
            <FileListDomain name="meshfile"/>
            <Documentation>This is the mesh file for a OcTree Mesh grid. This plugin only allows ONE mesh to be defined.</Documentation>
        </StringVectorProperty>''')
    def SetMeshFileName(self, fname):
        OcTreeReader.SetMeshFileName(self, fname)

    @smproperty.xml('''
        <StringVectorProperty
          panel_visibility="default"
          name="ModelFiles"
          label="File Name(s) Model"
          command="AddModelFileName"
          animateable="1"
          repeat_command="1"
          clean_command="ClearModels"
          number_of_elements="1">
          <FileListDomain name="modelfiles"/>
          <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
        </StringVectorProperty>''')
    def AddModelFileName(self, fname):
        """Use to set the file names for the reader. Handles singlt string or list of strings."""
        OcTreeReader.AddModelFileName(self, fname)


    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def SetTimeDelta(self, dt):
        OcTreeReader.SetTimeDelta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        """This is critical for registering the timesteps"""
        return OcTreeReader.GetTimestepValues(self)

    @smproperty.stringvector(name='DataName', default_values='Data')
    def SetDataName(self, name):
        OcTreeReader.SetDataName(self, name)



@smproxy.filter(name="PVGeoOcTreeAppender",
       label="Append Model To UBC OcTree Mesh")
@smhint.xml('''<RepresentationType view="RenderView" type="Surface With Edges" />''')
@smhint.xml('<ShowInMenu category="%s"/>' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkUnstructuredGrid"], composite_data_supported=False)
class PVGeoOcTreeAppender(OcTreeAppender):
    """This assumes the input vtkUnstructuredGrid has already handled the timesteps"""
    def __init__(self):
        OcTreeAppender.__init__(self)

    @smproperty.xml('''
        <StringVectorProperty
          panel_visibility="default"
          name="ModelFiles"
          label="File Name(s) Model"
          command="AddModelFileName"
          animateable="1"
          repeat_command="1"
          clean_command="ClearModels"
          number_of_elements="1">
          <FileListDomain name="modelfiles"/>
          <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
        </StringVectorProperty>''')
    def AddModelFileName(self, fname):
        """Use to set the file names for the reader. Handles single string or list of strings."""
        OcTreeAppender.AddModelFileName(self, fname)

    @smproperty.stringvector(name='DataName', default_values='Appended Data')
    def SetDataName(self, name):
        OcTreeAppender.SetDataName(self, name)


#------------------------------------------------------------------------------
# Write Tensor Mesh
#------------------------------------------------------------------------------
# @smproxy.writer(extensions="mesh", file_description="UBC Tensor Mesh", support_reload=False)
# @smproperty.input(name="Input", port_index=0)
# @smdomain.datatype(dataTypes=["vtkRectilinearGrid"], composite_data_supported=False)
# class vtkUBCTensorMeshWriter(VTKPythonAlgorithmBase):
#     def __init__(self):
#         VTKPythonAlgorithmBase.__init__(self, nInputPorts=1, nOutputPorts=0, inputType='vtkRectilinearGrid')
#         self._filename = None
#
#     @smproperty.stringvector(name="FileName", panel_visibility="never")
#     @smdomain.filelist()
#     def SetFileName(self, fname):
#         """Specify filename for the file to write."""
#         if self._filename != fname:
#             self._filename = fname
#             self.Modified()
#
#     def RequestData(self, request, inInfoVec, outInfoVec):
#         from vtkmodules.vtkCommonDataModel import vtkRectilinearGrid
#         from vtkmodules.numpy_interface import dataset_adapter as dsa
#
#         grid = dsa.WrapDataObject(vtkRectilinearGrid.GetData(inInfoVec[0], 0))
#         print('saving grid not impleneted')
#         return 1
#
#     def Write(self):
#         self.Modified()
#         self.Update()
#
# @smproxy.writer(extensions="gslib", file_description="GSLib Format", support_reload=False)
# @smproperty.input(name="Input", port_index=0)
# @smdomain.datatype(dataTypes=["vtkRectilinearGrid"], composite_data_supported=False)
# class vtkGSLIBMeshWriter(VTKPythonAlgorithmBase):
#     def __init__(self):
#         VTKPythonAlgorithmBase.__init__(self, nInputPorts=1, nOutputPorts=0, inputType='vtkRectilinearGrid')
#         self._filename = None
#
#     @smproperty.stringvector(name="FileName", panel_visibility="never")
#     @smdomain.filelist()
#     def SetFileName(self, fname):
#         """Specify filename for the file to write."""
#         if self._filename != fname:
#             self._filename = fname
#             self.Modified()
#
#     def RequestData(self, request, inInfoVec, outInfoVec):
#         #from vtkmodules.vtkCommonDataModel import vtkTable
#         #from vtkmodules.numpy_interface import dataset_adapter as dsa
#
#         table = dsa.WrapDataObject(vtkTable.GetData(inInfoVec[0], 0))
#         print('saving grid not impleneted')
#         return 1
#
#     def Write(self):
#         self.Modified()
#         self.Update()
#
#
# @smproxy.writer(extensions="bin", file_description="Binary Array (SEPLib/Madagascr)", support_reload=False)
# @smproperty.input(name="Input", port_index=0)
# @smdomain.datatype(dataTypes=["vtkRectilinearGrid"], composite_data_supported=False)
# class vtkBinaryWriter(VTKPythonAlgorithmBase):
#     def __init__(self):
#         VTKPythonAlgorithmBase.__init__(self, nInputPorts=1, nOutputPorts=0, inputType='vtkRectilinearGrid')
#         self._filename = None
#
#     @smproperty.stringvector(name="FileName", panel_visibility="never")
#     @smdomain.filelist()
#     def SetFileName(self, fname):
#         """Specify filename for the file to write."""
#         if self._filename != fname:
#             self._filename = fname
#             self.Modified()
#
#     def RequestData(self, request, inInfoVec, outInfoVec):
#         from vtkmodules.vtkCommonDataModel import vtkImageData
#         from vtkmodules.numpy_interface import dataset_adapter as dsa
#
#         grid = dsa.WrapDataObject(vtkImageData.GetData(inInfoVec[0], 0))
#         print('saving grid not impleneted')
#         return 1
#
#     def Write(self):
#         self.Modified()
#         self.Update()
